# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Contains datastore abstraction, used for saving study + trial data.

See resources.py for naming conventions.
"""
import abc
from typing import Callable, Iterable, List, Optional

from vizier._src.service import key_value_pb2
from vizier._src.service import resources
from vizier._src.service import study_pb2
from vizier._src.service import vizier_oss_pb2
from vizier._src.service import vizier_service_pb2
from google.longrunning import operations_pb2

UnitMetadataUpdate = vizier_service_pb2.UnitMetadataUpdate


class DataStore(abc.ABC):
  """Abstract class for data storage.

  Input/Outputs should always be pass-by-value.
  """

  @abc.abstractmethod
  def create_study(self, study: study_pb2.Study) -> resources.StudyResource:
    """Creates study in database. If preexisting, raises AlreadyExistsError."""

  @abc.abstractmethod
  def load_study(self, study_name: str) -> study_pb2.Study:
    """Loads study from the database. If nonexistent, raises NotFoundError."""

  @abc.abstractmethod
  def update_study(self, study: study_pb2.Study) -> resources.StudyResource:
    """Updates pre-existing study. If nonexistent, raises NotFoundError."""

  @abc.abstractmethod
  def delete_study(self, study_name: str) -> None:
    """Deletes study from database. If nonexistent, raises NotFoundError."""

  @abc.abstractmethod
  def list_studies(self, owner_name: str) -> List[study_pb2.Study]:
    """Lists all studies under owner.

    Args:
      owner_name: Name of owner.

    Returns:
      List of Studies associated with owner.

    Raises:
      NotFoundError: If the owner does not exist.
    """

  @abc.abstractmethod
  def create_trial(self, trial: study_pb2.Trial) -> resources.TrialResource:
    """Stores trial in database.

    Args:
      trial: Trial to be stored.

    Returns:
      The newly created trial resource name.

    Raises:
      AlreadyExistsError: If trial already exists.
    """

  @abc.abstractmethod
  def get_trial(self, trial_name: str) -> study_pb2.Trial:
    """Retrieves trial from database. If nonexistent, raises NotFoundError."""

  @abc.abstractmethod
  def update_trial(self, trial: study_pb2.Trial) -> resources.TrialResource:
    """Updates pre-existing trial. If nonexistent, raises NotFoundError."""

  @abc.abstractmethod
  def list_trials(self, study_name: str) -> List[study_pb2.Trial]:
    """List all trials for study. If study nonexistent, raises NotFoundError."""

  @abc.abstractmethod
  def delete_trial(self, trial_name: str) -> None:
    """Deletes trial from database. If nonexistent, raises NotFoundError."""

  @abc.abstractmethod
  def max_trial_id(self, study_name: str) -> int:
    """Maximal trial ID in study (defaults to 0 if no trials exist).

    Args:
      study_name: Name of study.

    Returns:
      Maximum trial ID as an int.

    Raises:
      NotFoundError: If the study does not exist.
    """

  @abc.abstractmethod
  def create_suggestion_operation(
      self, operation: operations_pb2.Operation
  ) -> resources.SuggestionOperationResource:
    """Stores suggestion operation.

    Args:
      operation:

    Returns:
      Resource for created op.

    Raises:
      AlreadyExistsError: If the suggest op already exists.
    """

  @abc.abstractmethod
  def get_suggestion_operation(
      self, operation_name: str
  ) -> operations_pb2.Operation:
    """Retrieves suggestion operation. If nonexistent, raises NotFoundError."""

  @abc.abstractmethod
  def update_suggestion_operation(
      self, operation: operations_pb2.Operation
  ) -> resources.SuggestionOperationResource:
    """Updates pre-existing suggestion op.

    Args:
      operation: New suggestion speration.

    Returns:
      Resource to operation.

    Raises:
      NotFoundError if suggest op is nonexistent.
    """

  @abc.abstractmethod
  def list_suggestion_operations(
      self,
      study_name: str,
      client_id: str,
      filter_fn: Optional[Callable[[operations_pb2.Operation], bool]] = None,
  ) -> List[operations_pb2.Operation]:
    """Retrieve all suggestion op from client.

    Args:
      study_name: Associated study for the suggest op.
      client_id: Associated client for the suggest op.
      filter_fn: Optional function to filter out the suggest ops.

    Raises:
      NotFoundError: If study or client is nonexistent.
    """

  @abc.abstractmethod
  def max_suggestion_operation_number(
      self, study_name: str, client_id: str
  ) -> int:
    """Maximal suggestion number for given client.

    Args:
      study_name: Name of associated owners.
      client_id: ID of associated client.

    Returns:
      Maximum suggest op number as an int.

    Raises:
      NotFoundError: If owner or client is nonexistent.
    """

  @abc.abstractmethod
  def create_early_stopping_operation(
      self, operation: vizier_oss_pb2.EarlyStoppingOperation
  ) -> resources.EarlyStoppingOperationResource:
    """Stores early stopping operation.

    Args:
      operation: Operation to be stored.

    Returns:
      The newly created op resource name.

    Raises:
      AlreadyExistsError: If the suggest op already exists.
    """

  @abc.abstractmethod
  def get_early_stopping_operation(
      self, operation_name: str
  ) -> vizier_oss_pb2.EarlyStoppingOperation:
    """Retrieves early stopping op. If nonexistent, raises NotFoundError."""

  @abc.abstractmethod
  def update_early_stopping_operation(
      self, operation: vizier_oss_pb2.EarlyStoppingOperation
  ) -> resources.EarlyStoppingOperationResource:
    """Updates preexisting early stopping op.

    Args:
      operation: New version of EarlyStoppingOp.

    Returns:
      Resource to the EarlyStoppingOp.

    Raises:
      NotFoundError: If early stopping op is nonexistent.
    """

  # TODO: Simplify the API by taking MetadataUpdateRequest proto
  # as input.
  @abc.abstractmethod
  def update_metadata(
      self,
      study_name: str,
      study_metadata: Iterable[key_value_pb2.KeyValue],
      trial_metadata: Iterable[UnitMetadataUpdate],
  ) -> None:
    """Store the supplied metadata in the database.

    Args:
      study_name: (Typically derived from a StudyResource.)
      study_metadata: Metadata to attach to the Study as a whole.
      trial_metadata: Metadata to attach to Trials.

    Raises:
      NotFoundError: If the update fails because of an attempt to attach
        metadata to a nonexistent Trial.
    """



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""HEBO GP models.


Faithful reimplementation of HEBO model based off of:
Paper: https://arxiv.org/abs/2012.03826
Repo: https://github.com/huawei-noah/HEBO.
"""

from typing import Generator, Optional
from flax import struct
import jax
from jax import numpy as jnp
from tensorflow_probability.substrates import jax as tfp
from vizier._src.jax import stochastic_process_model as sp
from vizier._src.jax import types
from vizier._src.jax.models import continuous_only_kernel

tfd = tfp.distributions
tfb = tfp.bijectors
tfpk = tfp.math.psd_kernels
tfpke = tfp.experimental.psd_kernels


@struct.dataclass
class VizierHeboGaussianProcess(sp.ModelCoroutine[tfd.GaussianProcess]):
  """Hebo Gaussian process model."""

  @classmethod
  def build_model(
      cls,
      features: types.ModelInput,
  ) -> sp.StochasticProcessModel:
    """Returns the model and loss function."""
    del features
    gp_coroutine = VizierHeboGaussianProcess()
    return sp.StochasticProcessModel(gp_coroutine)

  def __call__(
      self, inputs: Optional[types.ModelInput] = None
  ) -> Generator[sp.ModelParameter, jax.Array, tfd.GaussianProcess]:
    """Creates a generator.

    Args:
      inputs: tuple of (train_features, train_labels) train_features - array of
        dimension (num_examples, _feature_dim) train_labels - array of dimension
        (num_examples,)

    Yields:
      GaussianProcess whose event shape is `num_examples`.
    """
    epsilon = jnp.finfo(jnp.float64).resolution
    lim_val = jnp.float64(36.0)

    def _constraint_fn(x):
      return jnp.where(x > lim_val, x,
                       jnp.log1p(jnp.exp(jnp.clip(x, -lim_val,
                                                  lim_val)))) + epsilon

    def _inverse_constraint_fn(f):
      return jnp.where(f > lim_val, f, jnp.log(jnp.exp(f + 1e-20) - 1.))

    constraint_bijector = tfb.Inline(
        forward_fn=_constraint_fn,
        inverse_fn=_inverse_constraint_fn,
        forward_min_event_ndims=0)
    sigmoid = tfb.Sigmoid(jnp.float64(0.0), jnp.float64(10.0))

    # Signal variance
    gamma = tfd.Gamma(
        concentration=jnp.float64(0.5),
        rate=jnp.float64(1.0),
        name='signal_variance')
    signal_variance = yield sp.ModelParameter.from_prior(
        gamma,
        constraint=sp.Constraint((epsilon, None), bijector=constraint_bijector))

    # Observation noise variance
    noise_log_normal = tfd.LogNormal(
        loc=jnp.float64(-4.63), scale=0.5, name='observation_noise_variance')
    observation_noise_variance = yield sp.ModelParameter.from_prior(
        noise_log_normal,
        constraint=sp.Constraint((epsilon, None), bijector=constraint_bijector))

    # Kernel
    kernel = tfpk.MaternThreeHalves(
        amplitude=jnp.sqrt(signal_variance)) + tfpk.Linear()

    # Length scale
    length_scale = yield sp.ModelParameter.from_prior(
        tfd.LogNormal(
            loc=jnp.float64(0.0), scale=jnp.float64(1.0), name='length_scale'
        ),
        constraint=sp.Constraint((0.0, None), bijector=tfb.Exp()),
    )
    kernel = tfpk.FeatureScaled(kernel, scale_diag=length_scale)

    # Kumaraswamy input warping
    kumar_log_normal = tfd.LogNormal(loc=jnp.float64(0.0), scale=0.75)

    concentration0 = yield sp.ModelParameter.from_prior(
        kumar_log_normal.copy(name='concentration0'),
        constraint=sp.Constraint((0.0, 10.0), bijector=sigmoid),
    )

    concentration1 = yield sp.ModelParameter.from_prior(
        kumar_log_normal.copy(name='concentration1'),
        constraint=sp.Constraint((0.0, 10.0), bijector=sigmoid),
    )

    kernel = tfpk.KumaraswamyTransformed(kernel, concentration1, concentration0)

    kernel = continuous_only_kernel.ContinuousOnly(kernel)
    if inputs is not None:
      inputs = tfpke.ContinuousAndCategoricalValues(
          inputs.continuous.padded_array, inputs.categorical.padded_array
      )
    return tfd.GaussianProcess(
        kernel,
        index_points=inputs,
        observation_noise_variance=observation_noise_variance,
        cholesky_fn=None,
    )



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""NAS-Bench-101 Experimenter.

NOTE: Since our installation uses tensorflow>=2.0.0, you will need to change the
following in the nasbench package code (which assumes tensorflow-1 API):

```
tf.train.SessionRunHook -> tf.estimator.SessionRunHook
tf.train.CheckpointSaverListener -> tf.estimator.CheckpointSaverListener
tf.train.NanLossDuringTrainingError -> tf.estimator.NanLossDuringTrainingError
tf.python_io.tf_record_iterator -> tf.compat.v1.io.tf_record_iterator
```

See https://github.com/google-research/nasbench/issues/27.
"""
from typing import Sequence

from absl import logging
# from nasbench import api
import numpy as np

from vizier import pyvizier
from vizier._src.benchmarks.experimenters import experimenter

logging.warning(
    'NASBENCH-101 assumes tensorflow<2.0. Please see above in this file for what to replace in nasbench code to make the benchmark work.'
)


class NASBench101Experimenter(experimenter.Experimenter):
  """This suggests model specs in the form of a matrix (for DAG topology) and ops (for convolutions).

  Search space is a union of binary strings and categoricals.
  """

  def __init__(self, nasbench):
    self._nasbench = nasbench
    self._num_vertices = 7
    self._allowed_ops = ['conv3x3-bn-relu', 'conv1x1-bn-relu', 'maxpool3x3']
    self._op_spots = self._num_vertices - 2
    self._input_op = 'input'
    self._output_op = 'output'

    self._metric_names = [
        'trainable_parameters', 'training_time', 'train_accuracy',
        'validation_accuracy', 'test_accuracy'
    ]

  def evaluate(self, suggestions: Sequence[pyvizier.Trial]):
    """Evaluates and completes the Trial using NASBench-101 querying."""

    for trial in suggestions:
      spec = self._trial_to_model_spec(trial)
      if self._nasbench.is_valid(spec):
        results = self._nasbench.query(spec)
        trial.complete(
            pyvizier.Measurement(
                metrics={k: results[k] for k in self._metric_names}))
      else:
        trial.complete(
            pyvizier.Measurement(), infeasibility_reason='Not in search space.')

  def _trial_to_model_spec(self, trial: pyvizier.Trial):
    matrix = np.zeros((self._num_vertices, self._num_vertices), dtype=int)
    for y in range(self._num_vertices):
      for x in range(self._num_vertices):
        if y > x:
          matrix[x][y] = int(
              trial.parameters['{}_{}'.format(x, y)].value == 'True')

    base_ops = []
    for i in range(self._op_spots):
      base_ops.append(trial.parameters['ops_{}'.format(i)].value)
    ops = [self._input_op] + base_ops + [self._output_op]

    return api.ModelSpec(matrix=matrix, ops=ops)

  def problem_statement(self) -> pyvizier.ProblemStatement:
    problem_statement = pyvizier.ProblemStatement()
    root = problem_statement.search_space.root

    for y in range(self._num_vertices):
      for x in range(self._num_vertices):
        if y > x:
          root.add_bool_param(name='{}_{}'.format(x, y))

    for i in range(self._op_spots):
      root.add_categorical_param(
          name='ops_{}'.format(i), feasible_values=self._allowed_ops)

    # 'test_accuracy' also can be used for objective value.
    problem_statement.metric_information.append(
        pyvizier.MetricInformation(
            name='validation_accuracy',
            goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE))
    return problem_statement



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Convenience classes for configuring Vizier Early-Stopping Configs."""
import copy

import attr
from vizier._src.service import study_pb2

# When new early stopping config protos are added, include them below
# with a Union[]
AutomatedStoppingConfigProto = study_pb2.StudySpec.DefaultEarlyStoppingSpec


@attr.s(frozen=True, init=True, slots=True, kw_only=True)
class AutomatedStoppingConfig:
  """A wrapper for study_pb2.automated_stopping_spec."""
  _proto: AutomatedStoppingConfigProto = attr.ib(init=True, kw_only=True)

  @classmethod
  def default_stopping_spec(cls) -> 'AutomatedStoppingConfig':
    """Use Vizier's default early stopping."""
    config = study_pb2.StudySpec.DefaultEarlyStoppingSpec()
    return cls(proto=config)

  @classmethod
  def from_proto(
      cls, proto: AutomatedStoppingConfigProto) -> 'AutomatedStoppingConfig':
    return cls(proto=proto)

  def to_proto(self) -> AutomatedStoppingConfigProto:
    """Returns this object as a proto."""
    return copy.deepcopy(self._proto)



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Convenient types library for the /service/ folder."""
from typing import Union

from vizier._src.service import pythia_service_pb2_grpc
from vizier._src.service import vizier_service_pb2_grpc

# PythiaService and VizierService aliases allow us to use either a gRPC stub or
# the actual gRPC servicer implementation (for local setup) interchangeably.
PythiaService = Union[
    pythia_service_pb2_grpc.PythiaServiceStub,
    pythia_service_pb2_grpc.PythiaServiceServicer,
]

VizierService = Union[
    vizier_service_pb2_grpc.VizierServiceStub,
    vizier_service_pb2_grpc.VizierServiceServicer,
]



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for feature_mapper."""

from jax import config
import numpy as np
from vizier import pyvizier as vz
from vizier.pyvizier.converters import core
from vizier.pyvizier.converters import feature_mapper as fm

from absl.testing import absltest
from absl.testing import parameterized


class ContinuousCategoricalConverterTest(parameterized.TestCase):

  @parameterized.product(pad_oovs=[False, True])
  def test_categorical_only(self, pad_oovs):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_categorical_param('C1', ['a', 'b', 'c'])
    problem.search_space.root.add_categorical_param('C2', ['w', 'x', 'y', 'z'])
    problem.search_space.root.add_categorical_param('C3', ['0', '1'])
    problem.search_space.root.add_categorical_param('C4', ['*'])
    trial1 = vz.Trial(parameters={'C1': 'b', 'C2': 'z', 'C3': '0', 'C4': '*'})
    trial2 = vz.Trial(parameters={'C1': 'a', 'C2': 'z', 'C3': '0', 'C4': '*'})
    trial3 = vz.Trial(parameters={'C1': 'c', 'C2': 'x', 'C3': '1', 'C4': '*'})
    converter = core.TrialToArrayConverter.from_study_config(
        problem,
        pad_oovs=pad_oovs,
        max_discrete_indices=0,
    )
    features = converter.to_features([trial1, trial2, trial3])
    feature_mapper = fm.ContinuousCategoricalFeatureMapper(converter)
    res = feature_mapper.map(features)
    np.testing.assert_array_equal(
        res.categorical, np.array([[1, 3, 0, 0], [0, 3, 0, 0], [2, 1, 1, 0]])
    )
    np.testing.assert_array_equal(np.zeros((3, 0)), res.continuous)
    # Test un-mapping
    unmapped_features = feature_mapper.unmap(res)
    np.testing.assert_array_equal(unmapped_features, features)

  @parameterized.product(pad_oovs=[False, True], max_discrete_indices=[0, 10])
  def test_discrete_only(self, pad_oovs, max_discrete_indices):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_discrete_param('d1', [1, 10, 20])
    trial1 = vz.Trial(parameters={'d1': 10})
    trial2 = vz.Trial(parameters={'d1': 1})
    trial3 = vz.Trial(parameters={'d1': 20})
    converter = core.TrialToArrayConverter.from_study_config(
        problem, pad_oovs=pad_oovs, max_discrete_indices=max_discrete_indices
    )
    features = converter.to_features([trial1, trial2, trial3])
    feature_mapper = fm.ContinuousCategoricalFeatureMapper(converter)
    res = fm.ContinuousCategoricalFeatureMapper(converter).map(features)
    if max_discrete_indices == 10:
      # DISCRETE params are one-hot encoded. Should be mapped to categorical.
      np.testing.assert_array_equal(res.categorical, np.array([[1], [0], [2]]))
      np.testing.assert_array_equal(np.zeros((3, 0)), res.continuous)
    elif max_discrete_indices == 0:
      # DISCRETE params are continufied. Should be mapped to continuous.
      np.testing.assert_almost_equal(
          res.continuous, np.array([[(10 - 1) / (20 - 1)], [0.0], [1.0]])
      )
      np.testing.assert_array_equal(np.zeros((3, 0)), res.categorical)
    # Test un-mapping
    unmapped_features = feature_mapper.unmap(res)
    np.testing.assert_allclose(unmapped_features, features)

  def test_integer_only(self):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_int_param('i1', 0, 10)
    problem.search_space.root.add_int_param('i2', 0, 10)
    converter = core.TrialToArrayConverter.from_study_config(
        problem, pad_oovs=True, max_discrete_indices=0
    )
    trial1 = vz.Trial(parameters={'i1': 10, 'i2': 5})
    trial2 = vz.Trial(parameters={'i1': 1, 'i2': 3})
    trial3 = vz.Trial(parameters={'i1': 2, 'i2': 4})
    features = converter.to_features([trial1, trial2, trial3])
    feature_mapper = fm.ContinuousCategoricalFeatureMapper(converter)
    res = fm.ContinuousCategoricalFeatureMapper(converter).map(features)
    np.testing.assert_array_equal(
        res.continuous, np.array([[1.0, 0.5], [0.1, 0.3], [0.2, 0.4]])
    )
    np.testing.assert_array_equal(np.zeros((3, 0)), res.categorical)
    # Test un-mapping
    unmapped_features = feature_mapper.unmap(res)
    np.testing.assert_allclose(unmapped_features, features)

  def test_float_only(self):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('f1', 0.0, 1.0)
    problem.search_space.root.add_float_param('f2', 0.0, 1.0)
    converter = core.TrialToArrayConverter.from_study_config(
        problem, pad_oovs=True, max_discrete_indices=0
    )
    trial1 = vz.Trial(parameters={'f1': 1.0, 'f2': 0.5})
    trial2 = vz.Trial(parameters={'f1': 0.1, 'f2': 0.3})
    trial3 = vz.Trial(parameters={'f1': 0.2, 'f2': 0.4})
    features = converter.to_features([trial1, trial2, trial3])
    feature_mapper = fm.ContinuousCategoricalFeatureMapper(converter)
    res = fm.ContinuousCategoricalFeatureMapper(converter).map(features)
    np.testing.assert_allclose(
        res.continuous, np.array([[1.0, 0.5], [0.1, 0.3], [0.2, 0.4]])
    )
    np.testing.assert_array_equal(np.zeros((3, 0)), res.categorical)
    # Test un-mapping
    unmapped_features = feature_mapper.unmap(res)
    np.testing.assert_allclose(unmapped_features, features, atol=1e-5)

  @parameterized.product(pad_oovs=[False, True], max_discrete_indices=[0, 10])
  def test_mixed_space(self, pad_oovs, max_discrete_indices):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('f1', 0.0, 1.0)
    problem.search_space.root.add_float_param('f2', 0.0, 1.0)
    problem.search_space.root.add_discrete_param('d1', [1, 10, 20])
    problem.search_space.root.add_categorical_param('c1', ['a', 'b', 'c'])
    trial1 = vz.Trial(parameters={'f1': 1.0, 'f2': 0.5, 'd1': 10, 'c1': 'b'})
    trial2 = vz.Trial(parameters={'f1': 0.1, 'f2': 0.3, 'd1': 10, 'c1': 'a'})
    trial3 = vz.Trial(parameters={'f1': 0.2, 'f2': 0.4, 'd1': 20, 'c1': 'c'})
    converter = core.TrialToArrayConverter.from_study_config(
        problem, pad_oovs=pad_oovs, max_discrete_indices=max_discrete_indices
    )
    features = converter.to_features([trial1, trial2, trial3])
    feature_mapper = fm.ContinuousCategoricalFeatureMapper(converter)
    res = fm.ContinuousCategoricalFeatureMapper(converter).map(features)
    if max_discrete_indices == 10:
      np.testing.assert_array_equal(
          res.continuous, np.array([[1.0, 0.5], [0.1, 0.3], [0.2, 0.4]])
      )
      np.testing.assert_array_equal(
          res.categorical, np.array([[1, 1], [1, 0], [2, 2]])
      )
    if max_discrete_indices == 0:
      np.testing.assert_allclose(
          res.continuous,
          np.array([
              [1.0, 0.5, (10 - 1) / (20 - 1)],
              [0.1, 0.3, (10 - 1) / (20 - 1)],
              [0.2, 0.4, 1.0],
          ]),
      )
      np.testing.assert_array_equal(res.categorical, np.array([[1], [0], [2]]))
    # Test un-mapping
    unmapped_features = feature_mapper.unmap(res)
    np.testing.assert_allclose(unmapped_features, features)


if __name__ == '__main__':
  # Jax disables float64 computations by default and will silently convert
  # float64s to float32s. We must explicitly enable float64.
  config.update('jax_enable_x64', True)
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Constants."""

PARAMETER_NAME_ROOT = '$'
STUDY_METADATA_KEY_TUNER_ID = 'PRIMARY_TUNER_ID'
STUDY_METADATA_KEY_DNA_SPEC = 'TUNER_DNA_SPEC'
STUDY_METADATA_KEY_USES_EXTERNAL_DNA_SPEC = 'USES_EXTERNAL_DNA_SPEC'
TRIAL_METADATA_KEY_DNA_METADATA = 'DNA_METADATA'
TRIAL_METADATA_KEY_CUSTOM_TYPE_DECISIONS = 'CUSTOM_TYPE_DECISIONS'
TRIAL_METADATA_KEYS = frozenset(
    [TRIAL_METADATA_KEY_DNA_METADATA, TRIAL_METADATA_KEY_CUSTOM_TYPE_DECISIONS])
FROM_VIZIER_STUDY_HINT = 'from_vizier_study'
METADATA_NAMESPACE = 'pyglove'
RELATED_LINKS_SUBNAMESPACE = 'related_links'
DUMMY_PARAMETER_NAME = 'dummy_parameter'
DUMMY_PARAMETER_VALUE = 'UNUSED'

REWARD_METRIC_NAME = 'reward'

STUDY_METADATA_KEYS = frozenset([
    STUDY_METADATA_KEY_TUNER_ID,
    STUDY_METADATA_KEY_DNA_SPEC,
    STUDY_METADATA_KEY_USES_EXTERNAL_DNA_SPEC,
])



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for vizier.pyvizier.shared.base_study_config."""

import numpy as np
from vizier._src.pyvizier.shared import base_study_config
from absl.testing import absltest
from absl.testing import parameterized


class ObjectiveMetricGoalTest(absltest.TestCase):

  def test_basics(self):
    self.assertTrue(base_study_config.ObjectiveMetricGoal.MAXIMIZE.is_maximize)
    self.assertFalse(base_study_config.ObjectiveMetricGoal.MAXIMIZE.is_minimize)
    self.assertTrue(base_study_config.ObjectiveMetricGoal.MINIMIZE.is_minimize)
    self.assertFalse(base_study_config.ObjectiveMetricGoal.MINIMIZE.is_maximize)


class MetricTypeTest(absltest.TestCase):

  def test_basics(self):
    self.assertTrue(base_study_config.MetricType.SAFETY.is_safety)
    self.assertTrue(base_study_config.MetricType.OBJECTIVE.is_objective)


class MetricInformationTest(absltest.TestCase):

  def testMinMaxValueDefault(self):
    info = base_study_config.MetricInformation(
        goal=base_study_config.ObjectiveMetricGoal.MAXIMIZE)
    self.assertEqual(info.min_value, -np.inf)
    self.assertEqual(info.max_value, np.inf)

  def testMinMaxValueSet(self):
    info = base_study_config.MetricInformation(
        goal=base_study_config.ObjectiveMetricGoal.MAXIMIZE,
        min_value=-1.,
        max_value=1.)
    self.assertEqual(info.min_value, -1.)
    self.assertEqual(info.max_value, 1.)

  def testMinMaxBadValueInit(self):
    with self.assertRaises(ValueError):
      base_study_config.MetricInformation(
          goal=base_study_config.ObjectiveMetricGoal.MAXIMIZE,
          min_value=1.,
          max_value=-1.)

  def testMinMaxBadValueSet(self):
    info = base_study_config.MetricInformation(
        goal=base_study_config.ObjectiveMetricGoal.MAXIMIZE,
        min_value=-1.,
        max_value=1.)
    with self.assertRaises(ValueError):
      info.min_value = 2.
    with self.assertRaises(ValueError):
      info.max_value = -2.


class MetricsConfigTest(parameterized.TestCase):

  def testBasics(self):
    config = base_study_config.MetricsConfig()
    config.append(
        base_study_config.MetricInformation(
            name='max1', goal=base_study_config.ObjectiveMetricGoal.MAXIMIZE))
    config.extend([
        base_study_config.MetricInformation(
            name='max_safe1',
            goal=base_study_config.ObjectiveMetricGoal.MAXIMIZE,
            safety_threshold=0.0),
        base_study_config.MetricInformation(
            name='max2', goal=base_study_config.ObjectiveMetricGoal.MAXIMIZE),
        base_study_config.MetricInformation(
            name='min1', goal=base_study_config.ObjectiveMetricGoal.MINIMIZE),
        base_study_config.MetricInformation(
            name='min_safe2',
            goal=base_study_config.ObjectiveMetricGoal.MINIMIZE,
            safety_threshold=0.0,
            desired_min_safe_trials_fraction=0.1)
    ])
    self.assertLen(config, 5)
    self.assertLen(config.of_type(base_study_config.MetricType.OBJECTIVE), 3)
    self.assertLen(config.of_type(base_study_config.MetricType.SAFETY), 2)
    self.assertLen(
        config.exclude_type(base_study_config.MetricType.OBJECTIVE), 2
    )

  def testDuplicateNames(self):
    config = base_study_config.MetricsConfig()
    config.append(
        base_study_config.MetricInformation(
            name='max1', goal=base_study_config.ObjectiveMetricGoal.MAXIMIZE))
    with self.assertRaises(ValueError):
      config.append(
          base_study_config.MetricInformation(
              name='max1', goal=base_study_config.ObjectiveMetricGoal.MAXIMIZE))


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""HartMann 6-Dimensional function."""

from typing import Sequence
import numpy as np
from vizier import pyvizier as vz
from vizier._src.benchmarks.experimenters import experimenter
from vizier._src.benchmarks.experimenters import numpy_experimenter


def _hartmann6d(x: np.ndarray) -> float:
  """Hartmann function.

  Args:
    x: Shape (6,) array

  Returns:
    Shape (,) array.
  """
  # For math notations:
  # pylint: disable=invalid-name
  alpha = np.asarray([1.0, 1.2, 3.0, 3.2])
  A = np.asarray([
      [10, 3, 17, 3.5, 1.7, 8],
      [0.05, 10, 17, 0.1, 8, 14],
      [3, 3.5, 1.7, 10, 17, 8],
      [17, 8, 0.05, 10, 0.1, 14],
  ])
  P = 1e-4 * np.asarray([
      [1312, 1696, 5569, 124, 8283, 5886],
      [2329, 4135, 8307, 3736, 1004, 9991],
      [2348, 1451, 3522, 2883, 3047, 6650],
      [4047, 8828, 8732, 5743, 1091, 381],
  ])
  y = -alpha @ np.exp(-np.sum(A * (x - P) ** 2, axis=1))
  return y


class Hartmann6D(experimenter.Experimenter):
  """6D minimization function. See https://www.sfu.ca/~ssurjano/hart6.html."""

  def __init__(self):
    self._impl = numpy_experimenter.NumpyExperimenter(
        _hartmann6d, self.problem_statement()
    )

  def evaluate(self, suggestions: Sequence[vz.Trial]) -> None:
    self._impl.evaluate(suggestions)

  def problem_statement(self) -> vz.ProblemStatement:
    problem = vz.ProblemStatement()
    for i in range(1, 7):
      problem.search_space.root.add_float_param(f"x{i}", 0, 1)
    problem.metric_information.append(
        vz.MetricInformation(name="value", goal=vz.ObjectiveMetricGoal.MINIMIZE)
    )
    return problem


def _hartmann3d(x: np.ndarray) -> float:
  """Hartmann function.

  Args:
    x: Shape (6,) array

  Returns:
    Shape (,) array.
  """
  # For math notations:
  # pylint: disable=invalid-name
  alpha = np.asarray([1.0, 1.2, 3.0, 3.2])
  A = np.asarray([
      [3, 10, 30],
      [0.1, 10, 35],
      [3, 10, 30],
      [0.1, 10, 35],
  ])
  P = 1e-4 * np.asarray([
      [3689, 1170, 2673],
      [4699, 4387, 7470],
      [1091, 8732, 5547],
      [381, 5743, 8828],
  ])
  y = -alpha @ np.exp(-np.sum(A * (x - P) ** 2, axis=1))
  return y


class Hartmann3D(experimenter.Experimenter):
  """3D minimization function. See https://www.sfu.ca/~ssurjano/hart3.html."""

  def __init__(self):
    self._impl = numpy_experimenter.NumpyExperimenter(
        _hartmann3d, self.problem_statement()
    )

  def evaluate(self, suggestions: Sequence[vz.Trial]) -> None:
    self._impl.evaluate(suggestions)

  def problem_statement(self) -> vz.ProblemStatement:
    problem = vz.ProblemStatement()
    for i in range(1, 4):
      problem.search_space.root.add_float_param(f"x{i}", 0, 1)
    problem.metric_information.append(
        vz.MetricInformation(name="value", goal=vz.ObjectiveMetricGoal.MINIMIZE)
    )
    return problem



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for emukit."""

from vizier import pyvizier as vz
from vizier._src.algorithms.designers import emukit
from vizier._src.algorithms.testing import test_runners
from vizier.testing import test_studies

from absl.testing import absltest
from absl.testing import parameterized


class EmukitTest(parameterized.TestCase):

  @parameterized.parameters(((v,) for v in emukit.Version))
  def test_on_flat_space(self, version):
    config = vz.ProblemStatement(
        search_space=test_studies.flat_space_with_all_types(),
        metric_information=[
            vz.MetricInformation(
                name='x1', goal=vz.ObjectiveMetricGoal.MAXIMIZE)
        ])
    designer = emukit.EmukitDesigner(
        config, num_random_samples=10, metadata_ns='emukit', version=version)
    trials = test_runners.run_with_random_metrics(
        designer,
        config,
        iters=15,
        batch_size=1,
        verbose=1,
        validate_parameters=True)
    self.assertLen(trials, 15)

    for t in trials[:10]:
      self.assertEqual(
          t.metadata.ns('emukit')['source'],
          'random',
          msg=f'Trial {t} should be generated at random.')

    for t in trials[10:]:
      self.assertEqual(
          t.metadata.ns('emukit')['source'],
          'bayesopt',
          msg=f'Trial {t} should be generated by bayesopt.')


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Large-scale stress tests (multiple workers, multithreading) for PyGlove-Vizier integration."""
import multiprocessing.pool
import os
import random
import time
from absl import logging
import pyglove as pg

from vizier._src.pyglove import oss_vizier as vizier
from vizier._src.service import constants
from vizier._src.service import vizier_server

from absl.testing import absltest
from absl.testing import parameterized


NUM_TRIALS_PER_WORKER = 10
NUM_WORKERS = 10


class PerformanceTest(parameterized.TestCase):

  @classmethod
  def setUpClass(cls):
    super().setUpClass()
    server = vizier_server.DefaultVizierServer(
        host=os.uname()[1], database_url=constants.SQL_MEMORY_URL
    )
    logging.info(server.endpoint)
    vizier._services.reset_for_testing()
    vizier.init(vizier_endpoint=server.endpoint)
    cls.server = server
    logging.info('Vizier service is set up!')

  @parameterized.parameters(
      (multiprocessing.pool.ThreadPool,),
      # (multiprocessing.Pool,),  # Fails currently.
  )
  def test_multiple_workers(self, pool_creator):
    def work_fn(worker_id: int):
      del worker_id
      algorithm = pg.evolution.regularized_evolution()
      for _, feedback in pg.sample(
          pg.Dict(x=pg.oneof([1, 2, 3])),
          algorithm=algorithm,
          num_examples=NUM_TRIALS_PER_WORKER,
          name='performance_testing',
      ):
        feedback(reward=random.random())

    with pool_creator(NUM_WORKERS) as pool:
      start = time.time()
      pool.map(work_fn, range(NUM_WORKERS))
      end = time.time()

    logging.info(
        'For %d workers to evaluate %d trials each, it took %f seconds total.',
        NUM_WORKERS,
        NUM_TRIALS_PER_WORKER,
        end - start,
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for vizier.service.resources."""

from vizier._src.service import resources
from vizier._src.service import study_pb2
from vizier._src.service.testing import util as test_util

from absl.testing import absltest
from absl.testing import parameterized


class UtilTest(parameterized.TestCase):

  def setUp(self):
    self.owner_id = 'my_username'
    self.study_id = '12312312'
    self.trial_id = 1
    self.client_id = 'client_0'
    self.operation_number = 5
    super().setUp()

  def test_parsing_correct(self):
    """Tests for correct resource strings only."""
    owner_resource = resources.OwnerResource(self.owner_id)
    same_owner_resource = resources.OwnerResource.from_name(owner_resource.name)
    self.assertEqual(owner_resource, same_owner_resource)

    study_resource = resources.StudyResource(self.owner_id, self.study_id)
    same_study_resource = resources.StudyResource.from_name(study_resource.name)
    self.assertEqual(study_resource, same_study_resource)

    trial_resource = resources.TrialResource(
        self.owner_id, self.study_id, self.trial_id
    )
    same_trial_resource = resources.TrialResource.from_name(trial_resource.name)
    self.assertEqual(trial_resource, same_trial_resource)
    self.assertEqual(trial_resource.study_resource, study_resource)

    early_stopping_op_resource = resources.EarlyStoppingOperationResource(
        self.owner_id, self.study_id, self.trial_id
    )
    same_early_stopping_op_resource = (
        resources.EarlyStoppingOperationResource.from_name(
            early_stopping_op_resource.name
        )
    )
    self.assertEqual(
        early_stopping_op_resource, same_early_stopping_op_resource
    )
    self.assertEqual(early_stopping_op_resource.trial_resource, trial_resource)
    self.assertEqual(
        trial_resource.early_stopping_operation_resource,
        early_stopping_op_resource,
    )

    suggestion_op_resource = resources.SuggestionOperationResource(
        self.owner_id, self.study_id, self.client_id, self.operation_number
    )
    same_suggestion_op_resource = (
        resources.SuggestionOperationResource.from_name(
            suggestion_op_resource.name
        )
    )
    self.assertEqual(same_suggestion_op_resource, same_suggestion_op_resource)

  @parameterized.named_parameters(
      ('owner', 'owner/my_username', resources.OwnerResource),
      ('study', 'owners/my_username/study/cifar10', resources.StudyResource),
      (
          'trial',
          'owners/my_username/studies/cifar10/trials/not_an_int',
          resources.TrialResource,
      ),
  )
  def test_parsing_wrong(self, bad_name, resource_class):
    """Tests for incorrect resource strings, for input validation."""
    with self.assertRaises(ValueError):
      resource_class.from_name(bad_name)

  def test_generate_studies(self):
    study_spec = test_util.generate_all_four_parameter_specs()
    self.assertLen(study_spec.parameters, 4)

  def test_generate_trials(self):
    basic_trials = test_util.generate_trials(
        range(2), self.owner_id, self.study_id
    )
    self.assertLen(basic_trials, 2)
    all_states_trials = test_util.generate_all_states_trials(
        0, self.owner_id, self.study_id
    )
    self.assertLen(all_states_trials, len(study_pb2.Trial.State.keys()))

  def test_generate_suggestion_operations(self):
    basic_operations = test_util.generate_suggestion_operations(
        range(4), self.owner_id, self.study_id, self.client_id
    )
    self.assertLen(basic_operations, 4)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Acquisition functions and builders implementations."""

from typing import Callable, Mapping, Optional, Protocol

import chex
import equinox as eqx
from flax import struct
import jax
from jax import numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp
from vizier._src.algorithms.designers import scalarization
from vizier._src.jax import types
from vizier._src.jax.models import continuous_only_kernel


tfd = tfp.distributions
tfp_bo = tfp.experimental.bayesopt
tfpke = tfp.experimental.psd_kernels


class AcquisitionFunction(Protocol):
  """Acquisition function protocol."""

  # TODO: Acquisition functions should take xs as additional input.
  def __call__(
      self,
      dist: tfd.Distribution,
      seed: Optional[jax.Array] = None,
  ) -> jax.Array:
    pass


class ScoreFunction(Protocol):
  """Protocol for scoring candidate points."""

  def score(self, xs: types.ModelInput, seed: jax.Array) -> jax.Array:
    pass

  def score_with_aux(
      self, xs: types.ModelInput, seed: jax.Array
  ) -> tuple[jax.Array, chex.ArrayTree]:
    pass


class Predictive(Protocol):
  """Protocol for predicting distributions given candidate points."""

  def predict_with_aux(
      self, features: types.ModelInput
  ) -> tuple[tfd.Distribution, chex.ArrayTree]:
    pass


class ScoringFunctionFactory(Protocol):

  def __call__(
      self,
      data: types.ModelData,
      predictive: Predictive,
      use_trust_region: bool = False,
  ) -> ScoreFunction:
    pass


def sample_from_predictive(
    predictive: Predictive,
    xs: types.ModelInput,
    num_samples: int,
    *,
    key: jax.Array
) -> jax.Array:
  return predictive.predict_with_aux(xs)[0].sample([num_samples], seed=key)


def get_best_labels(labels: types.PaddedArray) -> jax.Array:
  """Returns the maximum values of labels.

  A note on "labels" in TFP acquisition functions: TFP acquisition functions
  (EI, PI, qEI, qUCB) take the maximum of `"observations"` (labels) over the
  rightmost axis, which is assumed to correspond to the number of observations.
  `best_labels` has a (singleton) rightmost dimension corresponding to the
  number of metrics. The shapes therefore work out correctly, although the
  semantics are different.

  Args:
    labels: Observed labels with padded shape `(num_observations, num_metrics)`.

  Returns: Maximum label values for each metric.
  """
  if jnp.size(labels.padded_array) == 0:
    return -np.inf
  return jnp.max(labels.replace_fill_value(-np.inf).padded_array, axis=-2)


def _apply_trust_region(
    region: 'TrustRegion',
    xs: types.ModelInput,
    acquisition: jax.Array,
    pred: tfd.Distribution,
    aux: chex.ArrayTree,
) -> tuple[jax.Array, chex.ArrayTree]:
  """Applies the trust region to acquisition values."""
  distance = region.min_linf_distance(xs)
  raw_acquisition = acquisition
  acquisition = jnp.where(
      ((distance <= region.trust_radius) | (region.trust_radius > 0.5)),
      acquisition,
      -1e12 - distance,
  )
  aux = aux | {
      'mean': pred.mean(),
      'stddev': pred.stddev(),
      'raw_acquisition': raw_acquisition,
      'linf_distance': distance,
      'radius': jnp.ones_like(distance) * region.trust_radius,
  }
  return acquisition, aux


class BayesianScoringFunction(eqx.Module):
  """Combines `Predictive` with acquisition function."""

  predictor: Predictive
  acquisition_fn: AcquisitionFunction

  # TODO: This should be moved out of here.
  # If set, uses trust region.
  trust_region: Optional['TrustRegion']

  def score(self, xs, seed: jax.Array) -> jax.Array:
    return self.score_with_aux(xs, seed)[0]

  def score_with_aux(
      self, xs, seed: jax.Array
  ) -> tuple[jax.Array, chex.ArrayTree]:
    jax.monitoring.record_event(
        '/vizier/jax/acquisitions/bayesian_scoring_function/score_with_aux/traced'
    )
    pred, aux = self.predictor.predict_with_aux(xs)

    acquisition = self.acquisition_fn(pred, seed=seed)
    if self.trust_region is not None:
      acquisition, aux = _apply_trust_region(
          self.trust_region,
          xs,
          acquisition,
          pred,
          aux,
      )
    return acquisition, aux


# Vizier library acquisition functions use `flax.struct`, instead of `attrs` and
# a hash function, so that acquisition functions can be passed as args to JIT-ed
# functions without triggering retracing when attribute values change.
@struct.dataclass
class UCB(AcquisitionFunction):
  """UCB AcquisitionFunction."""

  coefficient: float = 1.8

  def __call__(
      self,
      dist: tfd.Distribution,
      seed: Optional[jax.Array] = None,
  ) -> jax.Array:
    del seed
    return dist.mean() + self.coefficient * dist.stddev()


@struct.dataclass
class LCB(AcquisitionFunction):
  """LCB AcquisitionFunction."""

  coefficient: float = 1.8

  def __call__(
      self,
      dist: tfd.Distribution,
      seed: Optional[jax.Array] = None,
  ) -> jax.Array:
    del seed
    return dist.mean() - self.coefficient * dist.stddev()


@struct.dataclass
class HyperVolumeScalarization(AcquisitionFunction):
  """HyperVolume Scalarization acquisition function."""

  coefficient: float = 1.0

  def __call__(
      self,
      dist: tfd.Distribution,
      seed: Optional[jax.Array] = None,
  ) -> jax.Array:
    del seed
    # Uses scalarizations in https://arxiv.org/abs/2006.04655 for
    # non-convex biobjective optimization of mean vs stddev.
    return jnp.minimum(dist.mean(), self.coefficient * dist.stddev())


@struct.dataclass
class EI(AcquisitionFunction):
  """Expected Improvement acquisition function."""

  best_labels: jax.Array

  def __call__(
      self,
      dist: tfd.Distribution,
      seed: Optional[jax.Array] = None,
  ) -> jax.Array:
    del seed
    return tfp_bo.acquisition.GaussianProcessExpectedImprovement(
        dist, observations=self.best_labels
    )()


@struct.dataclass
class PI(AcquisitionFunction):
  """Probability of Improvement acquisition function."""

  best_labels: jax.Array

  def __call__(
      self,
      dist: tfd.Distribution,
      seed: Optional[jax.Array] = None,
  ) -> jax.Array:
    del seed
    return tfp_bo.acquisition.GaussianProcessProbabilityOfImprovement(
        dist, observations=self.best_labels
    )()


class MaxValueEntropySearch(eqx.Module):
  """MES score function. Implements the `ScoreFunction` protocol."""

  tfp_mes: tfp_bo.acquisition.GaussianProcessMaxValueEntropySearch
  trust_region: Optional['TrustRegion']

  @classmethod
  def scoring_fn_factory(
      cls,
      data: types.ModelData,
      predictive: Predictive,
      use_trust_region: bool = False,
  ) -> 'MaxValueEntropySearch':
    """Builds a MaxValueEntropySearch scoring function."""
    if data.features.categorical.shape[-1] > 0:
      raise ValueError('Categorical data is not supported.')
    pred_gp = predictive.predict_with_aux(data.features)[0]
    continuous_gp = pred_gp.copy(
        kernel=continuous_only_kernel.EmptyCategoricalKernel(pred_gp.kernel),
        observation_index_points=(pred_gp.observation_index_points.continuous),
        index_points=jnp.zeros_like(
            pred_gp.index_points.continuous,
            shape=(0,) + pred_gp.index_points.continuous.shape[1:],
        ),
        # pylint: disable=protected-access
        _conditional_kernel=continuous_only_kernel.EmptyCategoricalKernel(
            pred_gp._conditional_kernel
        ),
        _conditional_mean_fn=lambda x: pred_gp._conditional_mean_fn(  # pylint: disable=g-long-lambda
            continuous_only_kernel._continuous_to_cacv(x)
        ),
        # pylint: enable=protected-access
    )

    observations = pred_gp.observations
    # Allow the score function to build with no observations.
    if not observations.size:
      observations = jnp.asarray([-jnp.inf])
    tfp_mes = tfp_bo.acquisition.GaussianProcessMaxValueEntropySearch(
        predictive_distribution=continuous_gp,
        observations=observations,
        # TODO: Finalize API and plumb seed through.
        seed=jax.random.PRNGKey(123),
        num_max_value_samples=100,
    )
    trust_region = TrustRegion(data.features) if use_trust_region else None
    return cls(tfp_mes, trust_region=trust_region)

  def score(self, xs: types.ModelInput, seed: jax.Array) -> jax.Array:
    return self.score_with_aux(xs, seed)[0]

  def score_with_aux(
      self, xs: types.ModelInput, seed: jax.Array
  ) -> tuple[jax.Array, chex.ArrayTree]:
    del seed
    if xs.categorical.shape[-1] > 0:
      raise ValueError('Categorical data is not supported.')
    acquisition = self.tfp_mes(index_points=xs.continuous.padded_array)
    aux = {}
    if self.trust_region is not None:
      acquisition, aux = _apply_trust_region(
          self.trust_region,
          xs,
          acquisition,
          self.tfp_mes.predictive_distribution,
          aux,
      )
    return acquisition, aux


def bayesian_scoring_function_factory(
    acquisition_fn_factory: Callable[[types.ModelData], AcquisitionFunction],
) -> ScoringFunctionFactory:
  """Builds a ScoringFunctionFactory."""

  def f(
      data: types.ModelData,
      predictive: Predictive,
      use_trust_region: bool = False,
  ) -> ScoreFunction:
    acquisition_fn = acquisition_fn_factory(data)
    trust_region = TrustRegion(data.features) if use_trust_region else None
    return BayesianScoringFunction(predictive, acquisition_fn, trust_region)

  return f


@struct.dataclass
class AcquisitionTrustRegion(AcquisitionFunction):
  """Acquisition masked by a thresholding acquisition values.

  When optimizing the main acquisition, the goal is to consider idea is to
  disregard the region in the search space with
  thresholding_acquisition values lower than a threshold.
  Attributes:
    main_acquisition: The main acquisition function.
    thresholding_acquisition: The acquisition function used to detect promising
      (trust) regions.
    bad_acq_value: The lower bound for the main acquisition value.
    labels:
    threshold: The threshold for the thresholding_acquisition values to
      distinguish between the promising and unpromising regions.
    apply_tr_after: The minimum number of labels required to apply the trust
      region.
  """

  main_acquisition: AcquisitionFunction
  thresholding_acquisition: AcquisitionFunction
  bad_acq_value: float = struct.field(kw_only=True)
  labels: Optional[types.PaddedArray] = struct.field(kw_only=True)
  threshold: Optional[float] = struct.field(kw_only=True, default=None)
  apply_tr_after: Optional[int] = struct.field(kw_only=True, default=0)

  # TODO: Move factories to `vza_designer_factory.py`.
  @classmethod
  def default_ucb_pi(cls, data: types.ModelData) -> 'AcquisitionTrustRegion':
    best_labels = get_best_labels(data.labels)
    return cls(
        UCB(1.8),
        PI(best_labels),
        bad_acq_value=-1e12,
        labels=data.labels,
        threshold=0.3,
        apply_tr_after=0,
    )

  @classmethod
  def default_ucb_lcb(cls, data: types.ModelData) -> 'AcquisitionTrustRegion':
    return cls(
        UCB(1.8),
        LCB(1.8),
        labels=data.labels,
        bad_acq_value=-1e12,
        threshold=None,
        apply_tr_after=0,
    )

  @classmethod
  def default_ucb_lcb_wide(
      cls, data: types.ModelData
  ) -> 'AcquisitionTrustRegion':
    return cls(
        UCB(1.8),
        LCB(2.5),
        labels=data.labels,
        bad_acq_value=-1e12,
        threshold=None,
        apply_tr_after=0,
    )

  @classmethod
  def default_ucb_lcb_delay_tr(
      cls, data: types.ModelData
  ) -> 'AcquisitionTrustRegion':
    return cls(
        UCB(1.8),
        LCB(1.8),
        labels=data.labels,
        bad_acq_value=-1e12,
        threshold=None,
        apply_tr_after=5,
    )

  def __call__(
      self,
      dist: tfd.Distribution,
      seed: Optional[jax.Array] = None,
  ) -> jax.Array:
    del seed
    threshold_values = self.thresholding_acquisition(dist)
    acq_values = self.main_acquisition(dist)

    # TODO: Refactor so that the following is handled in the factory
    # instead.
    threshold = -jnp.inf
    apply_tr = False
    if self.labels is not None:
      labels_padded = self.labels.replace_fill_value(np.nan).padded_array
      threshold = jnp.minimum(
          jnp.nanmean(labels_padded), jnp.nanmedian(labels_padded)
      )
      apply_tr = self.labels._original_shape[0] <= self.apply_tr_after
    if self.threshold is not None:
      threshold = self.threshold
    cond = jnp.isnan(threshold) | (threshold_values >= threshold) | apply_tr
    return jnp.where(
        cond,
        acq_values,
        self.bad_acq_value - threshold_values,
    )


@struct.dataclass
class QEI(AcquisitionFunction):
  """Sampling-based batch expected improvement."""

  best_labels: jax.Array
  num_samples: int = 100

  def __call__(
      self,
      dist: tfd.Distribution,
      seed: Optional[jax.Array] = None,
  ) -> jax.Array:
    if seed is None:
      raise ValueError('QEI requires a value for `seed`.')
    return tfp_bo.acquisition.ParallelExpectedImprovement(
        dist,
        observations=self.best_labels,
        seed=seed,
        num_samples=self.num_samples,
    )()


@struct.dataclass
class QPI(AcquisitionFunction):
  """Sampling-based batch probability of improvement."""

  best_labels: jax.Array
  num_samples: int = 100

  def __call__(
      self,
      dist: tfd.Distribution,
      seed: Optional[jax.Array] = None,
  ) -> jax.Array:
    if seed is None:
      raise ValueError('QPI requires a value for `seed`.')
    return tfp_bo.acquisition.ParallelProbabilityOfImprovement(
        dist,
        observations=self.best_labels,
        seed=seed,
        num_samples=self.num_samples,
    )()


@struct.dataclass
class QUCB(AcquisitionFunction):
  """Sampling-based batch upper confidence bound.

  Attributes:
    coefficient: UCB coefficient. For a Gaussian distribution, note that
      `UCB(coefficient=c)` is equivalent to `QUCB(coefficient=c * sqrt(pi / 2))`
      if QUCB batch size is 1. See the TensorFlow Probability docs for more
      details:
      https://www.tensorflow.org/probability/api_docs/python/tfp/experimental/bayesopt/acquisition/ParallelUpperConfidenceBound
    num_samples: Number of distribution samples used to compute qUCB.
  """

  coefficient: float = 1.8
  num_samples: int = 100

  def __call__(
      self,
      dist: tfd.Distribution,
      seed: Optional[jax.Array] = None,
  ) -> jax.Array:
    if seed is None:
      raise ValueError('QEI requires a value for `seed`.')
    return tfp_bo.acquisition.ParallelUpperConfidenceBound(
        dist,
        seed=seed,
        exploration=self.coefficient,
        num_samples=self.num_samples,
        observations=None,
    )()


@struct.dataclass
class ScalarizedAcquisition(AcquisitionFunction):
  """Wrapper that scalarizes multiple objective before acquisition eval."""

  acquisition_fn: AcquisitionFunction
  scalarizer: scalarization.Scalarization

  def __call__(
      self,
      dist: tfd.Distribution,
      seed: Optional[jax.Array] = None,
  ) -> jax.Array:
    return jnp.squeeze(self.scalarizer(self.acquisition_fn(dist, seed)))


@struct.dataclass
class MultiAcquisitionFunction(AcquisitionFunction):
  """Wrapper that calls multiple acquisition functions."""

  acquisition_fns: Mapping[str, AcquisitionFunction]

  def __call__(
      self,
      dist: tfd.Distribution,
      seed: Optional[jax.Array] = None,
  ) -> jax.Array:
    acquisitions = []
    if seed is None:
      seeds = [None] * len(self.acquisition_fns)
    else:
      seeds = jax.random.split(seed, num=len(self.acquisition_fns))
    for i, acquisition_fn in enumerate(self.acquisition_fns.values()):
      acquisitions.append(acquisition_fn(dist, seeds[i]))
    # TODO: Change the return type to a dict with the same
    # structure as `acquisition_fns` to clarify the meaning of the return
    # values.
    return jnp.stack(acquisitions, axis=0)


# TODO: Support discretes and categoricals.
# TODO: Support custom distances.
class TrustRegion(eqx.Module):
  """L-inf norm based TrustRegion.

  Limits the suggestion within the union of small L-inf norm balls around each
  of the trusted points, which are in most cases observed points. The radius
  of the L-inf norm ball grows in the number of observed points.

  Assumes that all points are in the unit hypercube.

  The trust region can be used e.g. during acquisition optimization:
    converter = converters.TrialToArrayConverter.from_study_config(problem)
    features, labels = converter.to_xy(trials)
    tr = TrustRegion(features, converter.output_specs)
    # xs is a point in the search space.
    distance = tr.min_linf_distance(xs)
    if distance <= tr.trust_radius:
      print('xs in trust region')
  """
  trusted: types.ModelInput

  @property
  def trust_radius(self) -> jax.Array:
    # TODO: Make hyperparameters configurable.
    min_radius = 0.2  # Hyperparameter
    dimension_factor = 5.0  # Hyperparameter

    # pylint: disable=protected-access
    dof = (
        self.trusted.continuous._original_shape[-1]
        + self.trusted.categorical._original_shape[-1]
    )
    # pylint: enable=protected-access
    num_obs = jnp.sum(~self.trusted.continuous.is_missing[0])
    # TODO: Discount the infeasible points. The 0.1 and 0.9 split
    # is associated with weights to feasible and infeasbile trials.
    trust_level = (0.1 * num_obs + 0.9 * num_obs) / (
        dimension_factor * (dof + 1)
    )
    return jnp.where(
        num_obs == 0, 1.0, min_radius + (0.5 - min_radius) * trust_level
    )

  def min_linf_distance(self, xs: types.ModelInput) -> jax.Array:
    """l-inf norm distance to the closest trusted point.

    Caps distances between one-hot encoded features to the trust-region radius,
    so that the trust region cutoff does not discourage exploration of these
    features.

    Args:
      xs: (M, D) array where each element is in [0, 1].

    Returns:
      (M,) array of floating numbers, L-infinity distances to the nearest
      trusted point.
    """
    trusted = self.trusted.continuous.replace_fill_value(0.0).padded_array
    xs = xs.continuous.replace_fill_value(0.0).padded_array
    distances = jnp.abs(trusted - xs[..., jnp.newaxis, :])  # (M, N, D)
    # Mask out padded features. We set these distances to infinite since
    # they should never be considered.
    distances = jnp.where(
        self.trusted.continuous.is_missing[0][..., jnp.newaxis],
        np.inf,
        distances,
    )
    if distances.size == 0:
      return -np.inf * jnp.ones_like(xs, shape=xs.shape[:1])
    linf_distance = jnp.max(distances, axis=-1)  # (M, N)
    return jnp.min(linf_distance, axis=-1)  # (M,)



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Hard coded constants used in /service/ folder."""
import os

# The metadata namespace under which the Pythia endpoint is stored.
PYTHIA_ENDPOINT_NAMESPACE = 'service'

# The Study.metadata key where a Pythia endpoint can be stored.
PYTHIA_ENDPOINT_KEY = 'PYTHIA_ENDPOINT'

# A hard coded value which indicates that the Pythia endpoint is not set.
NO_ENDPOINT = 'NO_ENDPOINT'

# Default client ID.
UNUSED_CLIENT_ID = 'unused_client_id'

# Max int32 value.
MAX_STUDY_ID = 2147483647

# Will use RAM for SQL memory.
SQL_MEMORY_URL = 'sqlite:///:memory:'

# Will use local file path for HDD storage.
SERVICE_DIR = os.path.dirname(os.path.realpath(__file__))
VIZIER_DB_PATH = os.path.join(SERVICE_DIR, 'vizier.db')
SQL_LOCAL_URL = f'sqlite:///{VIZIER_DB_PATH}'



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for random_vectorized_optimizer."""

import numpy as np
from vizier import pyvizier as vz
from vizier._src.algorithms.optimizers import random_vectorized_optimizer as rvo
from vizier.pyvizier import converters

from absl.testing import absltest


class RandomVectorizedOptimizerTest(absltest.TestCase):
  """Tests for random_vectorized_optimizer."""

  def test_random_optimizer(self):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('f1', 0.0, 10.0)
    problem.search_space.root.add_float_param('f2', 0.0, 10.0)
    converter = converters.TrialToModelInputConverter.from_problem(problem)
    score_fn = lambda x, _: np.sum(x.continuous.padded_array, axis=(-1, -2))
    n_parallel = 3
    random_optimizer = rvo.create_random_optimizer(
        converter=converter, max_evaluations=100, suggestion_batch_size=10
    )
    res = random_optimizer(score_fn=score_fn, count=5, n_parallel=n_parallel)
    self.assertLen(res.rewards, 5)
    self.assertSequenceEqual(res.features.continuous.shape, (5, n_parallel, 2))

  def test_random_optimizer_factory(self):
    random_optimizer_factory = rvo.create_random_optimizer_factory(
        max_evaluations=100, suggestion_batch_size=10
    )
    self.assertEqual(random_optimizer_factory.suggestion_batch_size, 10)
    self.assertEqual(random_optimizer_factory.max_evaluations, 100)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Github test for run_tune."""
# pytype: skip-file
import itertools

import numpy as np
from ray import tune
from vizier._src.raytune import run_tune

from absl.testing import absltest


class RunTuneTest(absltest.TestCase):

  def test_simple_fit(self):
    # Uses random search by default.
    tune_config = tune.TuneConfig(num_samples=7, max_concurrent_trials=1)
    results = run_tune.run_tune_bbob(
        function_name='Sphere',
        dimension=5,
        shift=np.ones(5),
        tune_config=tune_config,
    )
    self.assertLen(results, 7)

  # TODO: Remove this when able to dedicate multi-threading.
  @absltest.skip('Parallelized fit fails due to multiple worker crashing.')
  def test_parallelized_fit(self):
    tune_config = tune.TuneConfig(num_samples=3)
    function_names = ['Sphere']
    dimensions = [4]
    product_list = list(itertools.product(function_names, dimensions))
    args_list = []
    for product in product_list:
      # Add args for shift, tune_config, run_config.
      args = product + (None, tune_config, None)
      args_list.append(args)
    results_list = run_tune.run_tune_distributed(
        args_list, run_tune.run_tune_bbob
    )

    # There should be 1 studies in the product, each with 3 Trials.
    self.assertLen(results_list, 1)
    for result in results_list:
      self.assertLen(result['result'], 3)

if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for vizier.pyvizier.shared.parameter_config."""

from typing import Any

from absl import logging
from vizier._src.pyvizier.shared import parameter_config as pc
from vizier._src.pyvizier.shared import trial
from vizier.testing import test_studies

from absl.testing import absltest
from absl.testing import parameterized


class ParameterConfigFactoryTest(parameterized.TestCase):

  def testCreatesDoubleConfig(self):
    parameter_config = pc.ParameterConfig.factory(
        'name',
        bounds=(-1.0, 1.0),
        scale_type=pc.ScaleType.LINEAR,
        default_value=0.1)
    self.assertEqual(parameter_config.name, 'name')
    self.assertEqual(parameter_config.type, pc.ParameterType.DOUBLE)
    self.assertEqual(parameter_config.bounds, (-1, 1))
    self.assertEqual(parameter_config.scale_type, pc.ScaleType.LINEAR)
    self.assertEqual(parameter_config.default_value, 0.1)
    self.assertIsInstance(parameter_config.default_value, float)
    with self.assertRaises(ValueError):
      _ = parameter_config.feasible_values

    self.assertEqual(parameter_config.continuify(), parameter_config)

  def testCreatesIntegerConfig(self):
    parameter_config = pc.ParameterConfig.factory(
        'name', bounds=(1, 3), scale_type=pc.ScaleType.LOG, default_value=1)
    self.assertEqual(parameter_config.name, 'name')
    self.assertEqual(parameter_config.type, pc.ParameterType.INTEGER)
    self.assertEqual(parameter_config.feasible_values, [1, 2, 3])
    self.assertEqual(parameter_config.bounds, (1, 3))
    self.assertEqual(parameter_config.scale_type, pc.ScaleType.LOG)
    self.assertEqual(parameter_config.default_value, 1)
    self.assertIsInstance(parameter_config.default_value, int)

    self.assertEqual(
        parameter_config.continuify(),
        pc.ParameterConfig.factory(
            'name',
            bounds=(1.0, 3.0),
            scale_type=pc.ScaleType.LOG,
            default_value=1.0))

  def testCreatesDiscreteConfig(self):
    feasible_values = (-1, 3, 2)
    parameter_config = pc.ParameterConfig.factory(
        'name',
        feasible_values=feasible_values,
        scale_type=pc.ScaleType.UNIFORM_DISCRETE,
        default_value=2,
        external_type=pc.ExternalType.INTEGER)
    self.assertEqual(parameter_config.name, 'name')
    self.assertEqual(parameter_config.type, pc.ParameterType.DISCRETE)
    self.assertEqual(parameter_config.feasible_values, [-1, 2, 3])
    self.assertEqual(parameter_config.bounds, (-1, 3))
    self.assertEqual(parameter_config.scale_type, pc.ScaleType.UNIFORM_DISCRETE)
    self.assertEqual(parameter_config.default_value, 2)
    self.assertIsInstance(parameter_config.default_value, float)
    self.assertEqual(parameter_config.external_type, pc.ExternalType.INTEGER)

    self.assertEqual(
        parameter_config.continuify(),
        pc.ParameterConfig.factory(
            'name', bounds=(-1.0, 3.0), default_value=2.0))

  def testCreatesCategoricalConfig(self):
    feasible_values = ('b', 'a', 'c')
    parameter_config = pc.ParameterConfig.factory(
        'name', feasible_values=feasible_values, default_value='c')
    self.assertEqual(parameter_config.name, 'name')
    self.assertEqual(parameter_config.feasible_values, ['a', 'b', 'c'])
    self.assertEqual(parameter_config.default_value, 'c')
    with self.assertRaises(ValueError):
      _ = parameter_config.bounds

  def testCreatesDoubleConfigIntDefault(self):
    parameter_config = pc.ParameterConfig.factory(
        'name',
        bounds=(-1.0, 1.0),
        scale_type=pc.ScaleType.LINEAR,
        default_value=1)
    self.assertEqual(parameter_config.default_value, 1.0)
    self.assertIsInstance(parameter_config.default_value, float)

  def testCreatesDiscreteConfigDoubleDefault(self):
    feasible_values = (-1, 3, 2)
    parameter_config = pc.ParameterConfig.factory(
        'name',
        feasible_values=feasible_values,
        scale_type=pc.ScaleType.UNIFORM_DISCRETE,
        default_value=2.0)
    self.assertEqual(parameter_config.default_value, 2.0)
    self.assertIsInstance(parameter_config.default_value, float)

  def testCreatesIntegerConfigDoubleDefault(self):
    parameter_config = pc.ParameterConfig.factory(
        'name', bounds=(1, 3), scale_type=pc.ScaleType.LOG, default_value=2.0)
    self.assertEqual(parameter_config.default_value, 2.0)
    self.assertIsInstance(parameter_config.default_value, int)

  def testCreatesIntegerConfigInvalidDoubleDefault(self):
    with self.assertRaisesRegex(ValueError, 'default_value for an.*'):
      pc.ParameterConfig.factory(
          'name',
          bounds=(1, 3),
          scale_type=pc.ScaleType.LOG,
          default_value=2.0001)

  def testCreatesCategoricalConfigNoDefault(self):
    feasible_values = ('b', 'a', 'c')
    parameter_config = pc.ParameterConfig.factory(
        'name', feasible_values=feasible_values)
    self.assertIsNone(parameter_config.default_value)

  def testCreatesCategoricalConfigBadDefault(self):
    feasible_values = ('b', 'a', 'c')
    with self.assertRaisesRegex(ValueError,
                                'default_value has an incorrect type.*'):
      pc.ParameterConfig.factory(
          'name', feasible_values=feasible_values, default_value=0.1)

  def testRaisesErrorWhenNameIsEmpty(self):
    with self.assertRaises(ValueError):
      _ = pc.ParameterConfig.factory('', bounds=(-1.0, 1.0))

  def testRaisesErrorWhenOverSpecified(self):
    with self.assertRaises(ValueError):
      _ = pc.ParameterConfig.factory(
          'name', bounds=(-1.0, 1.0), feasible_values=['a', 'b', 'c'])

  @parameterized.named_parameters(
      ('HaveInfinity', (-float('inf'), 1)), ('HaveNan', (1, float('nan'))),
      ('HaveMixedTypes', (1, float(1))), ('AreWronglyOrdered', (1, -1)))
  def testRaisesErrorWhenBounds(self, bounds):
    with self.assertRaises(ValueError):
      _ = pc.ParameterConfig.factory('name', bounds=bounds)

  @parameterized.named_parameters(('HaveDuplicateCategories', ['a', 'a', 'b']),
                                  ('HaveDuplicateNumbers', [1.0, 2.0, 2.0]),
                                  ('HaveMixedTypes', ['a', 1, 2]))
  def testRaisesErrorWhenFeasibleValues(self, feasible_values):
    with self.assertRaises(ValueError):
      _ = pc.ParameterConfig.factory('name', feasible_values=feasible_values)


_child1 = pc.ParameterConfig.factory('double_child', bounds=(0.0, 1.0))
_child2 = pc.ParameterConfig.factory('integer_child', bounds=(0, 1))


class ParameterConfigFactoryTestWithChildren(parameterized.TestCase):

  @parameterized.named_parameters(
      ('IntParentValues', [([0], _child1), ([0, 1], _child2)]),
      ('FloatParentValues', [([0.0], _child1), ([0.0, 1.0], _child2)]))
  def testIntegerWithValid(self, children):
    p = pc.ParameterConfig.factory('parent', bounds=(0, 1), children=children)
    self.assertCountEqual(p.subspace(0.0).parameters, [_child1, _child2])
    self.assertCountEqual(p.subspace(1.0).parameters, [_child2])

  @parameterized.named_parameters(
      ('FloatParentValues', [([0.5], _child1)]),
      ('StringParentValues', [(['0'], _child1), (['0.0', '1.0'], _child2)]))
  def testIntegerWithInvalid(self, children):
    with self.assertRaises(TypeError):
      _ = pc.ParameterConfig.factory('parent', bounds=(0, 1), children=children)

  @parameterized.named_parameters(
      ('IntParentValues', [([0], _child1), ([0, 1], _child2)]),
      ('FloatParentValues', [([0.0], _child1), ([0.0, 1.0], _child2)]))
  def testDiscreteWithValid(self, children):
    p = pc.ParameterConfig.factory(
        'parent', feasible_values=[0.0, 1.0], children=children)
    self.assertCountEqual(p.subspace(0.0).parameters, [_child1, _child2])
    self.assertCountEqual(p.subspace(1.0).parameters, [_child2])

  @parameterized.named_parameters(('StringParentValues', [(['0.0'], _child1),
                                                          (['0.0',
                                                            '1.0'], _child2)]))
  def testDiscreteWithInvalid(self, children):
    with self.assertRaises(TypeError):
      _ = pc.ParameterConfig.factory(
          'parent', feasible_values=[0.0, 1.0], children=children)

  @parameterized.named_parameters(  # pyformat: disable
      ('StringParentValues', [(['a'], _child1), (['a', 'b'], _child2)]))
  def testCategoricalWithValid(self, children):
    p = pc.ParameterConfig.factory(
        'parent', feasible_values=['a', 'b'], children=children)
    self.assertCountEqual(p.subspace('a').parameters, [_child1, _child2])
    self.assertCountEqual(p.subspace('b').parameters, [_child2])

  @parameterized.named_parameters(('StringParentValues', [(['0.0'], _child1),
                                                          (['1.0'], _child2)]))
  def testCategoricalWithInvalid(self, children):
    with self.assertRaises(TypeError):
      _ = pc.ParameterConfig.factory(
          'parent', feasible_values=[0.0, 1.0], children=children)


class MergeTest(parameterized.TestCase):

  def test_merge_bounds(self):
    pc1 = pc.ParameterConfig.factory('pc1', bounds=(0.0, 2.0))
    pc2 = pc.ParameterConfig.factory('pc2', bounds=(-1.0, 1.0))
    self.assertEqual(
        pc.ParameterConfig.merge(pc1, pc2),
        pc.ParameterConfig.factory('pc1', bounds=(-1.0, 2.0)))

  def test_merge_discrete(self):
    pc1 = pc.ParameterConfig.factory(
        'pc1', feasible_values=[0.0, 2.0], scale_type=pc.ScaleType.LINEAR)
    pc2 = pc.ParameterConfig.factory('pc2', feasible_values=[-1.0, 0.0])
    self.assertEqual(
        pc.ParameterConfig.merge(pc1, pc2),
        pc.ParameterConfig.factory(
            'pc1',
            feasible_values=[-1.0, 0.0, 2.0],
            scale_type=pc.ScaleType.LINEAR))

  def test_merge_categorical(self):
    pc1 = pc.ParameterConfig.factory('pc1', feasible_values=['a', 'b'])
    pc2 = pc.ParameterConfig.factory('pc2', feasible_values=['a', 'c'])
    self.assertEqual(
        pc.ParameterConfig.merge(pc1, pc2),
        pc.ParameterConfig.factory('pc1', feasible_values=['a', 'b', 'c']))


class ParameterConfigContainsTest(parameterized.TestCase):

  @parameterized.parameters((1.0, True), (-2.0, False), (3.0, False))
  def testFloat(self, value: Any, expected: bool):
    config = pc.ParameterConfig.factory('pc1', bounds=(-1., 2.))
    self.assertEqual(config.contains(value), expected)

  @parameterized.parameters((1, True), (-2, False), (3, False), (1.5, False))
  def testInt(self, value: Any, expected: bool):
    config = pc.ParameterConfig.factory('pc1', bounds=(-1, 2))
    self.assertEqual(config.contains(value), expected)

  @parameterized.parameters((1.0, False), (2, True), (-1, True))
  def testDiscrete(self, value: Any, expected: bool):
    config = pc.ParameterConfig.factory('pc1', feasible_values=[-1., 0., 2.])
    self.assertEqual(config.contains(value), expected)

  @parameterized.parameters(('a', True), ('b', False), ('c', False))
  def testCategorical(self, value: Any, expected: bool):
    config = pc.ParameterConfig.factory(
        'pc1', feasible_values=['a', 'aa', 'aaa'])
    self.assertEqual(config.contains(value), expected)

  @parameterized.parameters((True, True), ('a', False), (0, False))
  def testBoolean(self, value: Any, expected: bool):
    config = pc.ParameterConfig.factory(
        'pc1', feasible_values=[trial.TRUE_VALUE, trial.FALSE_VALUE])
    self.assertEqual(config.contains(value), expected)


class ParameterConfigPropertyTest(parameterized.TestCase):

  @parameterized.parameters(
      ((-1.0, 1.0), None), ((-1, 1), None), ((-2.0, -2.0), -2.0), ((-2, -2), -2)
  )
  def testFloatandInt(self, bounds: Any, expected: Any):
    config = pc.ParameterConfig.factory('pc1', bounds=bounds)
    value = config.deterministic_value
    if expected is None:
      self.assertIsNone(value)
    else:
      self.assertEqual(value, expected)

  @parameterized.parameters(
      ([-1.0, 2.0], None), ([-1.0], -1.0), (['a', 'b'], None), (['a'], 'a')
  )
  def testDiscreteandCategorical(self, feasible_values: Any, expected: Any):
    config = pc.ParameterConfig.factory('pc1', feasible_values=feasible_values)
    value = config.deterministic_value
    if expected is None:
      self.assertIsNone(value)
    else:
      self.assertEqual(value, expected)


class TraverseTest(parameterized.TestCase):

  @parameterized.named_parameters(('ShowChildrenTrue', True),
                                  ('ShowChildrenFalse', False))
  def testTraverse(self, show_children):
    grandchild1 = pc.ParameterConfig.factory('grandchild1', bounds=(-1.0, 1.0))
    grandchildren = [(['a'], grandchild1), (['b'], grandchild1)]
    child1 = pc.ParameterConfig.factory(
        'child1', feasible_values=['a', 'b'], children=grandchildren)

    child2 = pc.ParameterConfig.factory('child2', bounds=(0.0, 1.0))
    children = [([0], child1), ([1], child1), ([0, 1], child2)]
    parent = pc.ParameterConfig.factory(
        'parent', bounds=(0, 1), children=children)
    traversed_names = [
        pc.name for pc in parent.traverse(show_children=show_children)
    ]
    # Some parameter names are reused for separate child nodes, so they
    # will appear multiple times.
    self.assertEqual(traversed_names, [
        'parent', 'child1', 'grandchild1', 'grandchild1', 'child2', 'child1',
        'grandchild1', 'grandchild1', 'child2'
    ])


class SearchSpaceTest(parameterized.TestCase):
  """Check basic functionalities."""

  def testRootAndSelectRootEqual(self):
    space = pc.SearchSpace()
    select_root = space.select_root()
    shortcut_root = space.root
    self.assertIs(select_root._selected[0], shortcut_root._selected[0])

  def testAddFloatParamMinimal(self):
    # Remove this test once we deprecate select_root().
    space = pc.SearchSpace()
    self.assertEmpty(space.parameters)
    _ = space.select_root().add_float_param('f1', 1.0, 15.0)
    _ = space.select_root().add_float_param('f2', 2.0, 16.0)

    self.assertLen(space.parameters, 2)
    self.assertCountEqual(space.parameter_names, ['f1', 'f2'])

  def testMultidimensionalParameters(self):
    space = pc.SearchSpace()
    _ = space.select_root().add_float_param(
        'f', 1.0, 15.0, default_value=3.0, scale_type=pc.ScaleType.LOG, index=0)
    _ = space.select_root().add_float_param(
        'f',
        2.0,
        10.0,
        default_value=4.0,
        scale_type=pc.ScaleType.LINEAR,
        index=1)
    self.assertCountEqual(space.parameter_names, ['f[0]', 'f[1]'])


class SearchSpaceAddParamtest(parameterized.TestCase):
  """Check `add_xx_param` methods relay the arguments correctly."""

  def testAddFloatParam(self):
    space = pc.SearchSpace()
    self.assertEmpty(space.parameters)
    _ = space.select_root().add_float_param(
        'f1', 1.0, 15.0, default_value=3.0, scale_type=pc.ScaleType.LOG)
    self.assertLen(space.parameters, 1)
    self.assertEqual(space.parameters[0].name, 'f1')
    self.assertEqual(space.parameters[0].type, pc.ParameterType.DOUBLE)
    self.assertEqual(space.parameters[0].bounds, (1.0, 15.0))
    self.assertEqual(space.parameters[0].scale_type, pc.ScaleType.LOG)
    self.assertEmpty(space.parameters[0].matching_parent_values)
    self.assertEmpty(space.parameters[0].child_parameter_configs)
    with self.assertRaisesRegex(ValueError, 'feasible_values is invalid.*'):
      _ = space.parameters[0].feasible_values
    self.assertEqual(space.parameters[0].default_value, 3.0)

  def testAddDiscreteParamIntegerFeasibleValues(self):
    """Test a Discrete parameter with integer feasible values."""
    space = pc.SearchSpace()
    self.assertEmpty(space.parameters)
    _ = space.select_root().add_discrete_param(
        'd1', [101, 15.0, 21.0], default_value=15.0)
    self.assertLen(space.parameters, 1)
    self.assertEqual(space.parameters[0].name, 'd1')
    self.assertEqual(space.parameters[0].type, pc.ParameterType.DISCRETE)
    self.assertEqual(space.parameters[0].bounds, (15.0, 101.0))
    self.assertEqual(space.parameters[0].scale_type, pc.ScaleType.LINEAR)
    self.assertEmpty(space.parameters[0].matching_parent_values)
    self.assertEmpty(space.parameters[0].child_parameter_configs)
    self.assertEqual(space.parameters[0].feasible_values, [15.0, 21.0, 101])
    self.assertEqual(space.parameters[0].default_value, 15.0)
    self.assertEqual(space.parameters[0].external_type, pc.ExternalType.INTEGER)

  def testAddDiscreteParamFloatFeasibleValues(self):
    """Test a Discrete parameter with float feasible values."""
    space = pc.SearchSpace()
    self.assertEmpty(space.parameters)
    _ = space.select_root().add_discrete_param(
        'd1', [15.1, 21.0, 101], default_value=15.1)
    self.assertLen(space.parameters, 1)
    self.assertEqual(space.parameters[0].external_type, pc.ExternalType.FLOAT)

  def testAddBooleanParam(self):
    """Test a Boolean parameter."""
    space = pc.SearchSpace()
    self.assertEmpty(space.parameters)
    _ = space.select_root().add_bool_param('b1', default_value=True)
    self.assertLen(space.parameters, 1)
    self.assertEqual(space.parameters[0].name, 'b1')
    self.assertEqual(space.parameters[0].type, pc.ParameterType.CATEGORICAL)
    with self.assertRaisesRegex(ValueError,
                                'Accessing bounds of a categorical.*'):
      _ = space.parameters[0].bounds
    self.assertIsNone(space.parameters[0].scale_type)
    self.assertEmpty(space.parameters[0].matching_parent_values)
    self.assertEmpty(space.parameters[0].child_parameter_configs)
    self.assertEqual(space.parameters[0].feasible_values, ['False', 'True'])
    self.assertEqual(space.parameters[0].default_value, 'True')
    self.assertEqual(space.parameters[0].external_type, pc.ExternalType.BOOLEAN)

  def testAddBooleanParamWithFalseDefault(self):
    """Test a Boolean parameter."""
    space = pc.SearchSpace()
    self.assertEmpty(space.parameters)
    _ = space.select_root().add_bool_param('b1', default_value=False)
    self.assertLen(space.parameters, 1)
    self.assertEqual(space.parameters[0].default_value, 'False')

  def testAddCustomParam(self):
    """Test a Boolean parameter."""
    space = pc.SearchSpace()
    self.assertEmpty(space.parameters)
    _ = space.select_root().add_custom_param('c1', default_value='default')
    self.assertLen(space.parameters, 1)
    self.assertEqual(space.parameters[0].name, 'c1')
    self.assertEqual(space.parameters[0].default_value, 'default')

  def testConditionalParameters(self):
    space = pc.SearchSpace()
    self.assertEmpty(space.parameters)
    root = space.select_root()
    root.add_categorical_param(
        'model_type', ['linear', 'dnn'], default_value='dnn')

    # Test the search space.
    self.assertLen(space.parameters, 1)
    self.assertEqual(space.parameters[0].name, 'model_type')
    self.assertEqual(space.parameters[0].type, pc.ParameterType.CATEGORICAL)
    with self.assertRaisesRegex(ValueError,
                                'Accessing bounds of a categorical.*'):
      _ = space.parameters[0].bounds
    self.assertIsNone(space.parameters[0].scale_type)
    self.assertEmpty(space.parameters[0].matching_parent_values)
    self.assertEmpty(space.parameters[0].child_parameter_configs)
    self.assertEqual(space.parameters[0].feasible_values, ['dnn', 'linear'])
    self.assertEqual(space.parameters[0].default_value, 'dnn')

    dnn = root.select('model_type', ['dnn'])
    dnn.add_float_param(
        'learning_rate',
        0.0001,
        1.0,
        default_value=0.001,
        scale_type=pc.ScaleType.LOG)
    # Test the search space.
    self.assertLen(space.parameters, 1)
    self.assertLen(space.get('model_type').subspaces(), 1)

    linear = root.select('model_type', ['linear'])
    linear.add_float_param(
        'learning_rate',
        0.1,
        1.0,
        default_value=0.1,
        scale_type=pc.ScaleType.LOG)
    # Test the search space.
    self.assertLen(space.parameters, 1)
    self.assertLen(space.get('model_type').subspaces(), 2)

    _ = dnn.add_categorical_param('optimizer_type', ['adam', 'adagrad'])
    # Test the search space.
    self.assertLen(space.parameters, 1)

    # Chained select() calls, path length of 1.
    selected = root.select('model_type',
                           ['dnn']).select('optimizer_type',
                                           ['adam']).add_float_param(
                                               'learning_rate',
                                               0.1,
                                               1.0,
                                               default_value=0.1,
                                               scale_type=pc.ScaleType.LOG)
    self.assertLen(selected, 1)

    # Test the search space.
    self.assertLen(space.parameters, 1)

    # Chained select() calls, path length of 2.
    ko = root.select('model_type', ['dnn']).select('optimizer_type',
                                                   ['adam']).add_bool_param(
                                                       'use_keras_optimizer',
                                                       default_value=False)
    self.assertLen(ko, 1)
    # Test the search space.
    self.assertLen(space.parameters, 1)

    ko2 = ko.select_values(['True'])
    _ = ko2.add_float_param('keras specific', 1.3, 2.4, default_value=2.1)
    # Test the search space.
    self.assertLen(space.parameters, 1)

  def testConditionalParametersWithReturnedSelectors(self):
    space = pc.SearchSpace()
    self.assertEmpty(space.parameters)
    root = space.select_root()
    model_type = root.add_categorical_param('model_type', ['linear', 'dnn'])
    _ = model_type.select_values(['dnn']).add_float_param(
        'learning_rate',
        0.1,
        1.0,
        default_value=0.001,
        scale_type=pc.ScaleType.LOG)
    # Test the search space.
    self.assertLen(space.parameters, 1)

    # It is possible to select different values for the same selector.
    self.assertLen(
        model_type.select_values(['linear', 'dnn']).add_categorical_param(
            'optimizer_type', ['adam', 'adagrad']), 2)
    # Test the search space.
    self.assertLen(space.parameters, 1)

  @parameterized.named_parameters(
      ('Multi', 'units[0]', ('units', 0)),
      ('Multi2', 'with_underscore[1]', ('with_underscore', 1)),
      ('NotMulti', 'units', None),
      ('NotMulti2', 'with space', None),
      ('NotMulti3', 'with[8]space', None),
      ('NotMulti4', 'units[0][4]', ('units[0]', 4)),
      ('GinStyle', '_gin.ambient_net_exp_from_vec.block_type[3]',
       ('_gin.ambient_net_exp_from_vec.block_type', 3)),
  )
  def testParseMultiDimensionalParameterName(self, name, expected):
    base_name_index = pc.SearchSpaceSelector.parse_multi_dimensional_parameter_name(
        name)
    self.assertEqual(base_name_index, expected)

  def testValidateCategoricalInput(self):
    space = pc.SearchSpace()
    root = space.select_root()
    with self.assertRaises(ValueError):
      root.add_categorical_param('categorical', ['3.2', '2', 5])


class FlattenAndMergeTest(absltest.TestCase):

  def testFlattenAndMerge(self):
    space = test_studies.conditional_automl_space()
    parameters = space.root.select_all().merge()
    logging.info('Merged: %s', parameters)
    self.assertCountEqual(
        [p.name for p in parameters],
        [
            'model_type',
            'learning_rate',
            'optimizer_type',
            'use_special_logic',
            'special_logic_parameter',
        ],
    )


class SearchSpaceContainsTest(absltest.TestCase):

  def _space(self):
    space = pc.SearchSpace()
    root = space.select_root()
    root.add_float_param('learning-rate', 1e-4, 1e-2)
    root.add_categorical_param('optimizer', ['adagrad', 'adam', 'experimental'])
    return space

  def testFloatCat1(self):
    self._space().assert_contains(
        trial.ParameterDict({
            'optimizer': 'adagrad',
            'learning-rate': 1e-2
        }))

  def testFloatCat2(self):
    self.assertFalse(self._space().contains(
        trial.ParameterDict({
            'optimizer': 'adagrad',
            'BADPARAM': 1e-2
        })))

  def testFloatCat3(self):
    self.assertFalse(self._space().contains(
        trial.ParameterDict({
            'optimizer': 'adagrad',
            'learning-rate': 1e-2,
            'BADPARAM': 1e-2
        })))

  def testFloatCat4(self):
    self.assertFalse(self._space().contains(
        trial.ParameterDict({
            'optimizer': 'adagrad',
            'learning-rate': 1e2
        })))


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Import target for converters."""

from vizier.pyvizier.converters.core import DefaultModelInputConverter
from vizier.pyvizier.converters.core import DefaultModelOutputConverter
from vizier.pyvizier.converters.core import DefaultTrialConverter
from vizier.pyvizier.converters.core import dict_to_array
from vizier.pyvizier.converters.core import DictOf2DArrays
from vizier.pyvizier.converters.core import ModelInputConverter
from vizier.pyvizier.converters.core import ModelOutputConverter
from vizier.pyvizier.converters.core import NumpyArraySpec
from vizier.pyvizier.converters.core import NumpyArraySpecType
from vizier.pyvizier.converters.core import STUDY_ID_FIELD
from vizier.pyvizier.converters.core import TrialToArrayConverter
from vizier.pyvizier.converters.core import TrialToNumpyDict
from vizier.pyvizier.converters.embedder import ProblemAndTrialsScaler
from vizier.pyvizier.converters.jnp_converters import PaddedTrialToArrayConverter
from vizier.pyvizier.converters.jnp_converters import TrialToContinuousAndCategoricalConverter
from vizier.pyvizier.converters.jnp_converters import TrialToModelInputConverter
from vizier.pyvizier.converters.spatio_temporal import DenseSpatioTemporalConverter
from vizier.pyvizier.converters.spatio_temporal import SparseSpatioTemporalConverter
from vizier.pyvizier.converters.spatio_temporal import TimedLabels
from vizier.pyvizier.converters.spatio_temporal import TimedLabelsExtractor



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Test for converters. Cannot be tested internally but can be on GitHub."""
# pytype: skip-file
from ray import tune
from vizier import pyvizier as vz
from vizier._src.raytune import converters
from vizier.benchmarks import experimenters

from absl.testing import absltest


class ConvertersTest(absltest.TestCase):

  def test_to_dict(self):
    param_space = {
        'float': tune.uniform(1, 2),
        'float_log': tune.loguniform(2, 3),
        'int': tune.randint(4, 5),
        'choice': tune.choice(['a', 'b', 'c']),
    }
    search_space = converters.SearchSpaceConverter.to_vizier(param_space)
    self.assertIsInstance(search_space, vz.SearchSpace)
    self.assertLen(search_space.parameters, 4)

    config = search_space.get('float')
    self.assertEqual(config.type, vz.ParameterType.DOUBLE)
    self.assertEqual(config.bounds, (1, 2))
    self.assertEqual(config.scale_type, vz.ScaleType.LINEAR)

    config = search_space.get('float_log')
    self.assertEqual(config.type, vz.ParameterType.DOUBLE)
    self.assertEqual(config.bounds, (2, 3))
    self.assertEqual(config.scale_type, vz.ScaleType.LOG)

    config = search_space.get('int')
    self.assertEqual(config.type, vz.ParameterType.INTEGER)
    self.assertEqual(config.bounds, (4, 5))

    config = search_space.get('choice')
    self.assertEqual(config.type, vz.ParameterType.CATEGORICAL)
    self.assertEqual(config.feasible_values, ['a', 'b', 'c'])

  def test_run_study_with_search_space(self):
    space = vz.SearchSpace()
    root = space.select_root()
    root.add_float_param('uniform', 0.5, 1.5)
    root.add_float_param('loguniform', 0.1, 2.5, scale_type=vz.ScaleType.LOG)
    root.add_int_param('int_uniform', 1, 10)
    root.add_discrete_param('discrete', [1.0, 2.3, 0.5])
    root.add_categorical_param('categorical', ['a', 'b', 'c'])

    def trainable(config):  # Pass a "config" dictionary into your trainable.
      score = config['uniform'] * config['loguniform'] + config['int_uniform']
      if config['categorical'] == 'a':
        score += config['discrete']
      else:
        score -= config['discrete']

      return {'score': score}

    param_space = converters.SearchSpaceConverter.to_dict(space)
    tuner = tune.Tuner(
        trainable,
        param_space=param_space,
        tune_config=tune.TuneConfig(num_samples=10),
    )
    tuner.fit()
    self.assertLen(tuner.get_results(), 10)

  def test_conversion(self):
    dim = 2
    bbob_factory = experimenters.BBOBExperimenterFactory(name='Sphere', dim=dim)
    exptr = bbob_factory()
    search_space = exptr.problem_statement().search_space
    param_space = converters.SearchSpaceConverter.to_dict(search_space)
    trainable = converters.ExperimenterConverter.to_callable(exptr)

    config = {k: 1.5 for k in param_space.keys()}
    output_dict = trainable(config)
    self.assertEqual(type(output_dict['bbob_eval']), float)

  def test_run_study_with_experimenter(self):
    dim = 4
    bbob_factory = experimenters.BBOBExperimenterFactory(name='Sphere', dim=dim)
    exptr = bbob_factory()

    search_space = exptr.problem_statement().search_space
    self.assertLen(search_space.parameters, dim)

    param_space = converters.SearchSpaceConverter.to_dict(search_space)
    trainable = converters.ExperimenterConverter.to_callable(exptr)
    tuner = tune.Tuner(
        trainable,
        param_space=param_space,
        tune_config=tune.TuneConfig(num_samples=10),
    )
    tuner.fit()
    self.assertLen(tuner.get_results(), 10)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for vizier.service.vizier_server."""
# TODO: Change the test to create a vizier stub and call its
# methods, instead of directly calling VizierService methods.
import datetime
import time
import grpc

from vizier._src.service import constants
from vizier._src.service import key_value_pb2
from vizier._src.service import resources
from vizier._src.service import study_pb2
from vizier._src.service import vizier_service
from vizier._src.service import vizier_service_pb2
from vizier._src.service.testing import util as test_util

from google.longrunning import operations_pb2

from absl.testing import absltest
from absl.testing import parameterized

UnitMetadataUpdate = vizier_service_pb2.UnitMetadataUpdate


class VizierServicerTest(parameterized.TestCase):

  def setUp(self):
    # TODO: Find a way to cleanly test both datastores.
    self.early_stop_recycle_period = datetime.timedelta(seconds=0.1)
    self.vs = vizier_service.VizierServicer(
        database_url=constants.SQL_MEMORY_URL,
        early_stop_recycle_period=self.early_stop_recycle_period,
    )
    self.owner_id = 'my_username'
    self.study_id = '0123123'
    self.client_id = 'client_0'
    super().setUp()

  @parameterized.named_parameters(
      ('Maximize', study_pb2.StudySpec.MetricSpec.MAXIMIZE),
      ('Minimize', study_pb2.StudySpec.MetricSpec.MINIMIZE),
  )
  def test_single_objective_list_optimal(self, max_or_min_proto):
    metric_id = 'accuracy'
    max_or_min_study = test_util.generate_study(
        self.owner_id,
        self.study_id,
        study_spec=study_pb2.StudySpec(
            metrics=[
                study_pb2.StudySpec.MetricSpec(
                    metric_id=metric_id, goal=max_or_min_proto
                )
            ]
        ),
    )

    self.vs.datastore.create_study(max_or_min_study)

    lowest_trial_final_measurement = study_pb2.Measurement(
        metrics=[study_pb2.Measurement.Metric(metric_id=metric_id, value=-1.0)]
    )

    middle_trial_final_measurement = study_pb2.Measurement(
        metrics=[study_pb2.Measurement.Metric(metric_id=metric_id, value=0.0)]
    )

    highest_trial_final_measurement = study_pb2.Measurement(
        metrics=[study_pb2.Measurement.Metric(metric_id=metric_id, value=1.0)]
    )

    lowest_trial = test_util.generate_trials(
        trial_id_list=[1],
        owner_id=self.owner_id,
        study_id=self.study_id,
        final_measurement=lowest_trial_final_measurement,
        state=study_pb2.Trial.State.SUCCEEDED,
    )[0]

    middle_trial = test_util.generate_trials(
        trial_id_list=[2],
        owner_id=self.owner_id,
        study_id=self.study_id,
        final_measurement=middle_trial_final_measurement,
        state=study_pb2.Trial.State.SUCCEEDED,
    )[0]

    highest_trial = test_util.generate_trials(
        trial_id_list=[3],
        owner_id=self.owner_id,
        study_id=self.study_id,
        final_measurement=highest_trial_final_measurement,
        state=study_pb2.Trial.State.SUCCEEDED,
    )[0]

    another_highest_trial = test_util.generate_trials(
        trial_id_list=[4],
        owner_id=self.owner_id,
        study_id=self.study_id,
        final_measurement=highest_trial_final_measurement,
        state=study_pb2.Trial.State.SUCCEEDED,
    )[0]

    trials_to_be_ignored = test_util.generate_trials(
        [5, 6, 7], self.owner_id, self.study_id
    )

    all_trials = [
        lowest_trial,
        middle_trial,
        highest_trial,
        another_highest_trial,
    ] + trials_to_be_ignored

    for trial in all_trials:
      self.vs.datastore.create_trial(trial)

    optimal_trial_list = self.vs.ListOptimalTrials(
        request=vizier_service_pb2.ListOptimalTrialsRequest(
            parent=max_or_min_study.name
        )
    ).optimal_trials

    if max_or_min_study == study_pb2.StudySpec.MetricSpec.MAXIMIZE:
      self.assertLen(optimal_trial_list, 2)
    elif max_or_min_study == study_pb2.StudySpec.MetricSpec.MINIMIZE:
      self.assertEqual(optimal_trial_list[0], lowest_trial)

  def test_multiobjective_list_optimal(self):
    metric_id_1 = 'x1'
    metric_id_2 = 'x2'
    study = test_util.generate_study(
        self.owner_id,
        self.study_id,
        study_spec=study_pb2.StudySpec(
            metrics=[
                study_pb2.StudySpec.MetricSpec(
                    metric_id=metric_id_1,
                    goal=study_pb2.StudySpec.MetricSpec.MAXIMIZE,
                ),
                study_pb2.StudySpec.MetricSpec(
                    metric_id=metric_id_2,
                    goal=study_pb2.StudySpec.MetricSpec.MINIMIZE,
                ),
            ]
        ),
    )
    self.vs.datastore.create_study(study)

    low_x1 = study_pb2.Measurement.Metric(metric_id=metric_id_1, value=-1.0)
    low_x2 = study_pb2.Measurement.Metric(metric_id=metric_id_2, value=-1.0)
    high_x1 = study_pb2.Measurement.Metric(metric_id=metric_id_1, value=1.0)
    high_x2 = study_pb2.Measurement.Metric(metric_id=metric_id_2, value=1.0)

    low_low = study_pb2.Measurement(metrics=[low_x1, low_x2])
    low_high = study_pb2.Measurement(metrics=[low_x1, high_x2])
    high_low = study_pb2.Measurement(metrics=[high_x1, low_x2])
    high_high = study_pb2.Measurement(metrics=[high_x1, high_x2])

    incomplete_measurement = study_pb2.Measurement(metrics=[low_x1])

    ll_trial = test_util.generate_trials(
        trial_id_list=[1],
        owner_id=self.owner_id,
        study_id=self.study_id,
        final_measurement=low_low,
        state=study_pb2.Trial.State.SUCCEEDED,
    )[0]
    lh_trial = test_util.generate_trials(
        trial_id_list=[2],
        owner_id=self.owner_id,
        study_id=self.study_id,
        final_measurement=low_high,
        state=study_pb2.Trial.State.SUCCEEDED,
    )[0]
    hl_trial = test_util.generate_trials(
        trial_id_list=[3],
        owner_id=self.owner_id,
        study_id=self.study_id,
        final_measurement=high_low,
        state=study_pb2.Trial.State.SUCCEEDED,
    )[0]
    hh_trial = test_util.generate_trials(
        trial_id_list=[4],
        owner_id=self.owner_id,
        study_id=self.study_id,
        final_measurement=high_high,
        state=study_pb2.Trial.State.SUCCEEDED,
    )[0]

    missing_measurement_trials = test_util.generate_trials(
        trial_id_list=[5, 6, 7],
        owner_id=self.owner_id,
        study_id=self.study_id,
        final_measurement=incomplete_measurement,
        state=study_pb2.Trial.State.SUCCEEDED,
    )

    all_trials = [
        ll_trial,
        lh_trial,
        hl_trial,
        hh_trial,
    ] + missing_measurement_trials
    for trial in all_trials:
      self.vs.datastore.create_trial(trial)

    optimal_trial_list = self.vs.ListOptimalTrials(
        request=vizier_service_pb2.ListOptimalTrialsRequest(parent=study.name)
    ).optimal_trials

    self.assertLen(optimal_trial_list, 1)
    self.assertEqual(optimal_trial_list[0], hl_trial)

  def test_suggest_trials(self):
    suggestion_count = 10

    example_study_spec = test_util.generate_all_four_parameter_specs(
        algorithm='RANDOM_SEARCH'
    )
    example_study = test_util.generate_study(
        self.owner_id, self.study_id, study_spec=example_study_spec
    )
    self.vs.datastore.create_study(example_study)

    request = vizier_service_pb2.SuggestTrialsRequest(
        parent=resources.StudyResource(self.owner_id, self.study_id).name,
        suggestion_count=suggestion_count,
        client_id=self.client_id,
    )
    operation = self.vs.SuggestTrials(request)

    # Check if operation was stored in database.
    get_operation_request = operations_pb2.GetOperationRequest(
        name=resources.SuggestionOperationResource(
            self.owner_id, self.study_id, self.client_id, 1
        ).name
    )
    get_operation_output = self.vs.GetOperation(get_operation_request)
    self.assertEqual(operation, get_operation_output)

    # Check operation contents.
    suggest_trials_response = (
        vizier_service_pb2.SuggestTrialsResponse.FromString(
            operation.response.value
        )
    )
    self.assertLen(suggest_trials_response.trials, suggestion_count)
    for trial in suggest_trials_response.trials:
      self.assertLen(trial.parameters, 4)

    # Make sure a different op is created on a different study, even with same
    # client id.
    another_study_id = self.study_id + 'another'
    another_study = test_util.generate_study(
        self.owner_id, another_study_id, study_spec=example_study_spec
    )
    self.vs.datastore.create_study(another_study)
    another_request = vizier_service_pb2.SuggestTrialsRequest(
        parent=resources.StudyResource(self.owner_id, another_study_id).name,
        suggestion_count=suggestion_count,
        client_id=self.client_id,
    )
    another_operation = self.vs.SuggestTrials(another_request)
    self.assertNotEqual(operation, another_operation)

  @parameterized.named_parameters(
      ('IncludeFinalMeasurement', True), ('NoFinalMeasurement', False)
  )
  def test_complete_trial(self, include_final_measurement: bool):
    metric_id = 'x'
    study = test_util.generate_study(
        self.owner_id,
        self.study_id,
        study_spec=study_pb2.StudySpec(
            metrics=[
                study_pb2.StudySpec.MetricSpec(
                    metric_id=metric_id,
                    goal=study_pb2.StudySpec.MetricSpec.MAXIMIZE,
                )
            ]
        ),
    )
    self.vs.datastore.create_study(study)

    trial = test_util.generate_trials(
        trial_id_list=[1],
        owner_id=self.owner_id,
        study_id=self.study_id,
        state=study_pb2.Trial.State.ACTIVE,
    )[0]
    self.vs.datastore.create_trial(trial)
    complete_trial_request = vizier_service_pb2.CompleteTrialRequest(
        name=trial.name
    )

    if include_final_measurement:
      final_measurement = study_pb2.Measurement(
          metrics=[
              study_pb2.Measurement.Metric(metric_id=metric_id, value=-1.0)
          ]
      )
      complete_trial_request.final_measurement.CopyFrom(final_measurement)
      self.vs.CompleteTrial(complete_trial_request)
    else:
      with self.assertRaises(grpc.RpcError):
        # trial and request both do not contain measurements.
        self.vs.CompleteTrial(complete_trial_request)

  def test_early_stopping(self):
    example_study_spec = test_util.generate_all_four_parameter_specs(
        algorithm='RANDOM_SEARCH'
    )
    example_study = test_util.generate_study(
        self.owner_id, self.study_id, study_spec=example_study_spec
    )
    self.vs.datastore.create_study(example_study)

    active_trials = test_util.generate_trials(
        trial_id_list=[1, 2, 3, 4],
        owner_id=self.owner_id,
        study_id=self.study_id,
        state=study_pb2.Trial.State.ACTIVE,
    )
    for t in active_trials:
      self.vs.datastore.create_trial(t)
      request = vizier_service_pb2.CheckTrialEarlyStoppingStateRequest(
          trial_name=t.name
      )
      response = self.vs.CheckTrialEarlyStoppingState(request)
      # Since RandomPolicy picks a random ACTIVE trial to stop and current trial
      # t is the only ACTIVE trial, it should always stop.
      self.assertTrue(
          response.should_stop,
          msg=f'trial={t}, request={request}, response={response}',
      )
      stop_trial_request = vizier_service_pb2.StopTrialRequest(name=t.name)
      new_t = self.vs.StopTrial(stop_trial_request)
      self.assertEqual(new_t.state, study_pb2.Trial.State.STOPPING)

    for t in active_trials:
      trial_resource = resources.TrialResource.from_name(t.name)
      operation_name = resources.EarlyStoppingOperationResource(
          trial_resource.owner_id,
          trial_resource.study_id,
          trial_resource.trial_id,
      ).name
      op = self.vs.datastore.get_early_stopping_operation(operation_name)
      self.assertTrue(op.should_stop)

    # After a while, the opeartion will be recycled, and `should_stop` is
    # defaulted to False. Since the trial is no longer active, RandomPolicy will
    # not consider it.
    time.sleep(self.early_stop_recycle_period.total_seconds())
    request = vizier_service_pb2.CheckTrialEarlyStoppingStateRequest(
        trial_name=active_trials[0].name
    )
    response = self.vs.CheckTrialEarlyStoppingState(request)
    self.assertFalse(response.should_stop)

  def test_update_metadata(self):
    # Construct a study.
    example_study_spec = test_util.generate_all_four_parameter_specs(
        algorithm='RANDOM_SEARCH'
    )
    example_study = test_util.generate_study(
        self.owner_id, self.study_id, study_spec=example_study_spec
    )
    self.vs.datastore.create_study(example_study)
    active_trials = test_util.generate_trials(
        trial_id_list=[1, 2],
        owner_id=self.owner_id,
        study_id=self.study_id,
        state=study_pb2.Trial.State.ACTIVE,
    )
    for t in active_trials:
      self.vs.datastore.create_trial(t)

    # Construct the request.
    study_metadata = UnitMetadataUpdate(
        metadatum=key_value_pb2.KeyValue(key='a', ns='b', value='C')
    )
    trial_metadata = UnitMetadataUpdate(
        trial_id='1',
        metadatum=key_value_pb2.KeyValue(key='d', ns='e', value='F'),
    )
    request = vizier_service_pb2.UpdateMetadataRequest(
        name=resources.StudyResource(self.owner_id, self.study_id).name,
        delta=[study_metadata, trial_metadata],
    )
    # Send it to the server.
    response = self.vs.UpdateMetadata(request)
    # Check that there was no error.
    self.assertEmpty(response.error_details)

  def test_trial_immutable(self):
    study = test_util.generate_study(
        self.owner_id, self.study_id, state=study_pb2.Study.State.ACTIVE
    )
    self.vs.datastore.create_study(study)
    trials = test_util.generate_trials(
        trial_id_list=[1],
        owner_id=self.owner_id,
        study_id=self.study_id,
        state=study_pb2.Trial.State.SUCCEEDED,
    )
    self.vs.datastore.create_trial(trials[0])
    trial_name = resources.TrialResource(self.owner_id, self.study_id, 1).name

    with self.assertRaises(grpc.RpcError):
      self.vs.CompleteTrial(
          vizier_service_pb2.CompleteTrialRequest(
              name=trial_name, final_measurement=study_pb2.Measurement()
          )
      )

    with self.assertRaises(grpc.RpcError):
      self.vs.CheckTrialEarlyStoppingState(
          vizier_service_pb2.CheckTrialEarlyStoppingStateRequest(
              trial_name=trial_name
          )
      )

    with self.assertRaises(grpc.RpcError):
      self.vs.AddTrialMeasurement(
          vizier_service_pb2.AddTrialMeasurementRequest(
              trial_name=trial_name, measurement=study_pb2.Measurement()
          )
      )

  @parameterized.parameters(
      (study_pb2.Study.State.COMPLETED,),
      (study_pb2.Study.State.INACTIVE,),
  )
  def test_study_immutable(self, study_state):
    trial_name = resources.TrialResource(self.owner_id, self.study_id, 1).name
    study_name = resources.StudyResource(self.owner_id, self.study_id).name
    study = test_util.generate_study(
        self.owner_id, self.study_id, state=study_state
    )
    self.vs.datastore.create_study(study)

    with self.assertRaises(grpc.RpcError):
      self.vs.SuggestTrials(
          vizier_service_pb2.SuggestTrialsRequest(parent=study_name)
      )
    with self.assertRaises(grpc.RpcError):
      self.vs.CreateTrial(
          vizier_service_pb2.CreateTrialRequest(
              parent=study_name, trial=study_pb2.Trial()
          )
      )
    with self.assertRaises(grpc.RpcError):
      self.vs.AddTrialMeasurement(
          vizier_service_pb2.AddTrialMeasurementRequest(
              trial_name=trial_name, measurement=study_pb2.Measurement()
          )
      )
    with self.assertRaises(grpc.RpcError):
      self.vs.CompleteTrial(
          vizier_service_pb2.CompleteTrialRequest(name=trial_name)
      )
    with self.assertRaises(grpc.RpcError):
      self.vs.DeleteTrial(
          vizier_service_pb2.DeleteTrialRequest(name=trial_name)
      )
    with self.assertRaises(grpc.RpcError):
      self.vs.CheckTrialEarlyStoppingState(
          vizier_service_pb2.CheckTrialEarlyStoppingStateRequest(
              trial_name=trial_name
          )
      )
    with self.assertRaises(grpc.RpcError):
      self.vs.StopTrial(vizier_service_pb2.StopTrialRequest(name=trial_name))
    with self.assertRaises(grpc.RpcError):
      self.vs.UpdateMetadata(
          vizier_service_pb2.UpdateMetadataRequest(name=study_name)
      )


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for base_runner."""

from vizier import pyvizier as vz
from vizier._src.algorithms.designers import random
from vizier._src.benchmarks.experimenters import experimenter_factory
from vizier._src.benchmarks.runners import benchmark_runner
from vizier._src.benchmarks.runners import benchmark_state

from absl.testing import absltest
from absl.testing import parameterized


def _get_designer_benchmark_state_factory():
  dim = 10
  experimenter = experimenter_factory.BBOBExperimenterFactory('Sphere', dim)()

  def _designer_factory(config: vz.ProblemStatement, seed: int):
    return random.RandomDesigner(config.search_space, seed=seed)

  benchmark_state_factory = benchmark_state.DesignerBenchmarkStateFactory(
      designer_factory=_designer_factory, experimenter=experimenter
  )

  return benchmark_state_factory


class BaseRunnerTest(parameterized.TestCase):

  @parameterized.parameters(
      {
          'runner': benchmark_runner.BenchmarkRunner(
              benchmark_subroutines=[
                  benchmark_runner.GenerateSuggestions(),
                  benchmark_runner.EvaluateActiveTrials(),
              ],
              num_repeats=7,
          ),
          'expected_trials': 7,
      },
      {
          'runner': benchmark_runner.BenchmarkRunner(
              benchmark_subroutines=[benchmark_runner.GenerateAndEvaluate(10)],
              num_repeats=5,
          ),
          'expected_trials': 50,
      },
      {
          'runner': benchmark_runner.BenchmarkRunner(
              benchmark_subroutines=[
                  benchmark_runner.FillActiveTrials(10),
                  benchmark_runner.EvaluateActiveTrials(),
              ],
              num_repeats=5,
          ),
          'expected_trials': 50,
      },
  )
  def test_benchmark_run(self, runner, expected_trials):
    benchmark_state_factory = _get_designer_benchmark_state_factory()
    bench_state = benchmark_state_factory(seed=5)
    runner.run(bench_state)
    self.assertEmpty(
        bench_state.algorithm.supporter.GetTrials(
            status_matches=vz.TrialStatus.ACTIVE
        )
    )
    all_trials = bench_state.algorithm.supporter.GetTrials()
    self.assertLen(all_trials, expected_trials)
    for trial in all_trials:
      self.assertEqual(trial.status, vz.TrialStatus.COMPLETED)

  def test_active_trials(self):
    benchmark_state_factory = _get_designer_benchmark_state_factory()
    bench_state = benchmark_state_factory(seed=5)
    runner = benchmark_runner.BenchmarkRunner(
        benchmark_subroutines=[
            benchmark_runner.GenerateSuggestions(10),
            benchmark_runner.EvaluateActiveTrials(6),
        ],
        num_repeats=3,
    )
    runner.run(bench_state)
    self.assertLen(
        bench_state.algorithm.supporter.GetTrials(
            status_matches=vz.TrialStatus.ACTIVE
        ),
        4 * 3,
    )
    self.assertLen(
        bench_state.algorithm.supporter.GetTrials(
            status_matches=vz.TrialStatus.COMPLETED
        ),
        6 * 3,
    )

  def test_fill_active_trials(self):
    benchmark_state_factory = _get_designer_benchmark_state_factory()
    bench_state = benchmark_state_factory(seed=5)
    runner = benchmark_runner.BenchmarkRunner(
        benchmark_subroutines=[
            benchmark_runner.FillActiveTrials(10),
            benchmark_runner.EvaluateActiveTrials(6),  # 6 Completions
            benchmark_runner.FillActiveTrials(3),  # No-op
            benchmark_runner.EvaluateActiveTrials(6),  # 4 Completions
            benchmark_runner.FillActiveTrials(10),
            benchmark_runner.EvaluateActiveTrials(6),  # 6 completions
        ]
    )
    runner.run(bench_state)
    self.assertLen(
        bench_state.algorithm.supporter.GetTrials(
            status_matches=vz.TrialStatus.ACTIVE
        ),
        4,
    )
    self.assertLen(
        bench_state.algorithm.supporter.GetTrials(
            status_matches=vz.TrialStatus.COMPLETED
        ),
        16,
    )

  def test_benchmark_run_from_exptr_factory(self):
    benchmark_state_factory = _get_designer_benchmark_state_factory()
    bench_state = benchmark_state_factory(seed=5)
    runner = benchmark_runner.BenchmarkRunner(
        benchmark_subroutines=[benchmark_runner.GenerateAndEvaluate(10)],
        num_repeats=5,
    )
    runner.run(bench_state)
    self.assertEmpty(
        bench_state.algorithm.supporter.GetTrials(
            status_matches=vz.TrialStatus.ACTIVE
        )
    )
    all_trials = bench_state.algorithm.supporter.GetTrials()
    self.assertLen(all_trials, 50)
    for trial in all_trials:
      self.assertEqual(trial.status, vz.TrialStatus.COMPLETED)

  def test_add_prior(self):
    dim = 10
    exptr_factory = experimenter_factory.BBOBExperimenterFactory('Discus', dim)

    def _designer_factory(config: vz.ProblemStatement, seed: int):
      return random.RandomDesigner(config.search_space, seed=seed)

    prior_benchmark_state_factory = (
        benchmark_state.ExperimenterDesignerBenchmarkStateFactory(
            designer_factory=_designer_factory,
            experimenter_factory=exptr_factory,
        )
    )
    prior_runner = benchmark_runner.BenchmarkRunner(
        benchmark_subroutines=[benchmark_runner.GenerateAndEvaluate(10)],
        num_repeats=5,
    )
    prior_study_guid = 'prior'
    runner = benchmark_runner.EvaluateAndAddPriorStudy(
        benchmark_runner=prior_runner,
        benchmark_state_factory=prior_benchmark_state_factory,
        study_guid=prior_study_guid,
    )

    benchmark_state_factory = _get_designer_benchmark_state_factory()
    bench_state = benchmark_state_factory(seed=5)
    runner.run(bench_state)
    self.assertEmpty(bench_state.algorithm.supporter.GetTrials())

    prior_trials = bench_state.algorithm.supporter.GetTrials(
        study_guid=prior_study_guid
    )
    self.assertLen(prior_trials, 50)
    for trial in prior_trials:
      self.assertEqual(trial.status, vz.TrialStatus.COMPLETED)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Acquisition functions and builders implementations."""

from typing import Optional

import jax
from jax import numpy as jnp
from tensorflow_probability.substrates import jax as tfp
from vizier._src.algorithms.designers.gp import acquisitions as acquisitions_lib
from vizier._src.jax import stochastic_process_model as sp
from vizier._src.jax import types


tfd = tfp.distributions
tfp_bo = tfp.experimental.bayesopt
tfpke = tfp.experimental.psd_kernels


def _build_predictive_distribution(
    xs: types.ModelInput,
    model: sp.StochasticProcessModel,
    state: types.GPState,
    use_vmap: bool = True,
) -> tfd.Distribution:
  """Generates the predictive distribution on array function."""

  def _predict_on_array_one_model(
      model_state: types.ModelState, *, xs: types.ModelInput
  ) -> tfd.Distribution:
    return model.apply(
        model_state,
        xs,
        state.data.features,
        state.data.labels,
        method=model.posterior_predictive,
    )

  if not use_vmap:
    return _predict_on_array_one_model(state.model_state, xs=xs)

  def _predict_mean_and_stddev(state_: types.ModelState) -> tfd.Distribution:
    dist = _predict_on_array_one_model(state_, xs=xs)
    return {'mean': dist.mean(), 'stddev': dist.stddev()}  # pytype: disable=attribute-error  # numpy-scalars

  # Returns a dictionary with mean and stddev, of shape [M, N].
  # M is the size of the parameter ensemble and N is the number of points.
  pp = jax.vmap(_predict_mean_and_stddev)(state.model_state)
  batched_normal = tfd.Normal(pp['mean'].T, pp['stddev'].T)  # pytype: disable=attribute-error  # numpy-scalars

  return tfd.MixtureSameFamily(
      tfd.Categorical(logits=jnp.ones(batched_normal.batch_shape[1])),
      batched_normal,
  )


def predict_on_array(
    xs: types.ModelInput,
    model: sp.StochasticProcessModel,
    state: types.GPState,
    use_vmap: bool = True,
):
  """Prediction function on features array."""
  dist = _build_predictive_distribution(xs, model, state, use_vmap)
  return {'mean': dist.mean(), 'stddev': dist.stddev()}


def acquisition_on_array(
    xs: types.ModelInput,
    model: sp.StochasticProcessModel,
    acquisition_fn: acquisitions_lib.AcquisitionFunction,
    state: types.GPState,
    trust_region: Optional[acquisitions_lib.TrustRegion] = None,
    use_vmap: bool = True,
):
  """Acquisition function on features array."""
  dist = _build_predictive_distribution(xs, model, state, use_vmap)
  acquisition = acquisition_fn(dist)
  if trust_region is not None:
    distance = trust_region.min_linf_distance(xs)
    # Due to output normalization, acquisition can't be nearly as
    # low as -1e12.
    # We use a bad value that decreases in the distance to trust region
    # so that acquisition optimizer can follow the gradient and escape
    # untrusted regions.
    return jnp.where(
        (trust_region.trust_radius >= 0.5)
        | (
            (distance <= trust_region.trust_radius)
            & (trust_region.trust_radius < 0.5)
        ),
        acquisition,
        -1e12 - distance,
    )
  else:
    return acquisition



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""TunerPolicy."""

from typing import Optional, Sequence, cast

from absl import logging
import attr
import pyglove as pg
from vizier import pythia
from vizier import pyvizier as vz
from vizier._src.algorithms.policies import trial_caches
from vizier._src.pyglove import constants
from vizier._src.pyglove import converters
from vizier._src.pyglove import core


@attr.define
class TunerPolicy(pythia.Policy):
  """Pythia policy for custom multi-trial tuner algorithm.

  Note that study_config should be used if the user needs the Trials to
  faithfully use the parameter names from the original StudyConfig.
  """

  supporter: pythia.PolicySupporter = attr.field()
  _converter: converters.VizierConverter = attr.field()
  algorithm: pg.geno.DNAGenerator = attr.field()
  early_stopping_policy: Optional[pg.tuning.EarlyStoppingPolicy] = attr.field(
      default=None
  )

  def __attrs_post_init__(self):
    # Initialize the caches. Unlike Vizier's Pythia service implementation,
    # Pyglove's pythia service maintains a single Policy instance for each
    # study (unless the binary crashes). We can therefore use an in-ram cache
    # and avoid re-loading the same trials over and over again.
    self._suggestion_cache = trial_caches.IdDeduplicatingTrialLoader(
        self.supporter, include_intermediate_measurements=False
    )
    self._stopping_cache = trial_caches.IdDeduplicatingTrialLoader(
        self.supporter, include_intermediate_measurements=True
    )

  @property
  def _metric_names(self) -> Sequence[str]:
    return self._converter.metrics_to_optimize

  def _update(self, tuner_trial: pg.tuning.Trial) -> bool:
    """Update a single tuner Trial.

    Args:
      tuner_trial: If the trial id was previously seen, update is no-op.

    Returns:
      True if the trial was added.
    """
    logging.info(
        'Updating TunerTrial id=%s to algorithm: %s',
        tuner_trial.id,
        self.algorithm,
    )
    reward = tuner_trial.get_reward_for_feedback(self._metric_names)
    if reward is not None:
      self.algorithm.feedback(tuner_trial.dna, reward)
      return True
    return False

  def suggest(self, request: pythia.SuggestRequest) -> pythia.SuggestDecision:
    logging.info('TunerPolicy:suggest started')
    newly_completed_trials = self._suggestion_cache.get_newly_completed_trials(
        request.max_trial_id
    )
    n_trials_updated = 0
    metadata_updates = vz.MetadataDelta()
    for vizier_trial in newly_completed_trials:
      tuner_trial = core.VizierTrial(self._converter, vizier_trial)
      metadata = dict(tuner_trial.dna.metadata)  # make a deep copy.
      if self._update(tuner_trial):
        n_trials_updated += 1
        # Serialize DNA metadata to Vizier trial metadata upon change.
        if pg.ne(tuner_trial.dna.metadata, metadata):
          metadata_updates.assign(
              namespace=constants.METADATA_NAMESPACE,
              key=constants.TRIAL_METADATA_KEY_DNA_METADATA,
              value=pg.to_json_str(tuner_trial.dna.metadata),
              trial=vizier_trial,
          )

    new_trials = []
    count = request.count or 1

    logging.info(
        (
            'The algorithm is updated with %s new trials out of %s newly'
            ' completed trials. It has seen %s trials total. Now generating %s'
            ' new suggestions.'
        ),
        n_trials_updated,
        len(newly_completed_trials),
        self._suggestion_cache.num_incorporated_trials,
        count,
    )
    for _ in range(request.count or 1):
      try:
        dna = self.algorithm.propose()
      except StopIteration:
        logging.info(
            'Search algorithm %s has no new suggestions.', self.algorithm
        )
        break

      if dna.spec is None:
        dna.use_spec(self._converter.dna_spec)
      trial = self._converter.to_trial(dna, fallback='return_dummy')
      new_trials.append(trial)
      logging.info(
          'algorithm : %s proposed DNA: %s which is translated to trial: %s',
          self.algorithm,
          repr(dna),
          trial,
      )

    logging.info('TunerPolicy:suggest ended')
    return pythia.SuggestDecision(new_trials, metadata_updates)

  def early_stop(
      self, request: pythia.EarlyStopRequest
  ) -> pythia.EarlyStopDecisions:
    if self.early_stopping_policy is None:
      return pythia.EarlyStopDecisions()

    early_stopping_policy = cast(
        pg.tuning.EarlyStoppingPolicy, self.early_stopping_policy
    )

    newly_completed_trials = self._stopping_cache.get_newly_completed_trials(
        request.max_trial_id
    )
    for vizier_trial in newly_completed_trials:
      tuner_trial = core.VizierTrial(self._converter, vizier_trial)
      # For completed trials, we do not need the actual stopping decision but
      # we still have to feed them in and update the stopping policy's state.
      early_stopping_policy.should_stop_early(tuner_trial)

    active_trials = self._stopping_cache.get_active_trials()

    decisions = pythia.EarlyStopDecisions()
    for vizier_trial in active_trials:
      tuner_trial = core.VizierTrial(self._converter, vizier_trial)
      should_stop = early_stopping_policy.should_stop_early(tuner_trial)
      decisions.decisions.append(
          pythia.EarlyStopDecision(
              vizier_trial.id,
              should_stop=should_stop,
              reason='Pyglove stopping policy stopped the trial.',
          )
      )
    return decisions


def create_policy(
    supporter: pythia.PolicySupporter,
    problem_statement: vz.ProblemStatement,
    algorithm: pg.geno.DNAGenerator,
    early_stopping_policy: Optional[pg.tuning.EarlyStoppingPolicy] = None,
    prior_trials: Optional[Sequence[vz.Trial]] = None,
) -> pythia.Policy:
  """Creates a Pythia policy that uses PyGlove algorithms."""
  converter = converters.VizierConverter.from_problem(problem_statement)

  # Bind the algorithm with the search space before usage.
  algorithm.setup(converter.dna_spec)

  # Warm up algorithm if prior trials are present.
  if prior_trials:

    def get_trial_history():
      for trial in prior_trials:
        tuner_trial = core.VizierTrial(converter, trial)
        reward = tuner_trial.get_reward_for_feedback(
            converter.metrics_to_optimize
        )
        yield (tuner_trial.dna, reward)

    algorithm.recover(get_trial_history())

  return TunerPolicy(supporter, converter, algorithm, early_stopping_policy)
